"""Bag-of-words logistic regression."""

import numpy as np
import tensorflow as tf

from rayleaf.models.model import Model
from utils.data_utils import batch_data


class ClientModel(Model):

    def __init__(self, seed, lr, num_classes, input_dim):
        self.num_classes = num_classes
        self.input_dim = input_dim
        super(ClientModel, self).__init__(seed, lr)

    def create_model(self):
        features = tf.placeholder(tf.float32, [None, self.input_dim])
        labels = tf.placeholder(tf.int64, [None])

        logits = tf.layers.dense(features, self.num_classes, activation=tf.nn.sigmoid)
        loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=labels,
            logits=logits)
        
        train_op = self.optimizer.minimize(
            loss=loss,
            global_step=tf.train.get_global_step())

        predictions = tf.argmax(logits, axis=-1)
        correct_pred = tf.equal(predictions, labels)
        eval_metric_ops = tf.count_nonzero(correct_pred)
        
        return features, labels, train_op, eval_metric_ops, tf.reduce_mean(loss)

    def process_x(self, raw_x_batch):
        return np.array(raw_x_batch)

    def process_y(self, raw_y_batch):
        return np.array(raw_y_batch)

    def _run_epoch(self, data, batch_size):
        for batched_x, batched_y in batch_data(data, batch_size, self.seed):
            input_data = self.process_x(batched_x)
            target_data = self.process_y(batched_y)

            with self.graph.as_default():
                self.sess.run(
                    self.train_op,
                    feed_dict={
                        self.features: input_data,
                        self.labels: target_data})

    def _test(self, data):
        x_vecs = self.process_x(data['x'])
        labels = self.process_y(data['y'])

        with self.graph.as_default():
            tot_acc, loss = self.sess.run(
                [self.eval_metric_ops, self.loss],
                feed_dict={
                    self.features: x_vecs, 
                    self.labels: labels
                })

        acc = float(tot_acc) / len(x_vecs)
        return {'accuracy': acc}



import pandas as pd

from tabulate import tabulate


logging_file = None


def log(msg: str = "") -> None:
    global logging_file

    print(msg, file=logging_file)
    print(msg)


def log_df(df: pd.DataFrame):
    log(tabulate(df, headers="keys", tablefmt="psql", showindex=False))


def shutdown_log() -> None:
    global logging_file
    
    if logging_file is not None:
        logging_file.close()
    logging_file = None



def online(clients: list) -> list:
    """We assume all users are always online."""
    return sorted(clients.keys())


def count_selected_clients(selected_clients: list, clients: dict) -> dict:
    counts = {}
    for client_num in selected_clients:
        client_type = clients[client_num]
        counts[client_type] = counts.get(client_type, 0) + 1

    return counts


def client_counts_string(counts: dict) -> str:
    client_types = list(counts.keys())
    client_types.sort(key=lambda ClientType: ClientType.__name__)

    counts_str = ""
    count_format = "{count} {client_type_name}"

    if len(client_types) == 1:
        counts_str += count_format.format(count=counts[client_types[0]], client_type_name=client_types[0].__name__)
        if counts[client_types[0]] != 1:
            counts_str += "s"
    elif len(client_types) == 2:
        counts_str += count_format.format(count=counts[client_types[0]], client_type_name=client_types[0].__name__)
        if counts[client_types[0]] != 1:
            counts_str += "s"

        counts_str += " and "

        counts_str += count_format.format(count=counts[client_types[1]], client_type_name=client_types[1].__name__)
        if counts[client_types[1]] != 1:
            counts_str += "s"
    else:
        for client_type in client_types[:-1]:
            counts_str += count_format.format(count=counts[client_type], client_type_name=client_type.__name__)
            if counts[client_type] != 1:
                counts_str += "s"
            counts_str += ", "
        
        counts_str += "and "
        counts_str += count_format.format(count=counts[client_types[-1]], client_type_name=client_types[-1].__name__)
        if counts[client_types[-1]] != 1:
            counts_str += "s"
    
    return counts_str



import torch
import numpy as np

from torch import Tensor


import rayleaf


def test_simple():
    shape1 = (3, 1, 4, 1)
    shape2 = (1, 5, 9)

    arr1 = np.ones(shape1)
    arr2 = np.ones(shape2)
    arr = [arr1, arr2]

    ta = rayleaf.TensorArray(arr)
    flat_ta = Tensor(np.ones((np.prod(shape1) + np.prod(shape2))))

    assert torch.all(ta.__array__() == flat_ta)



import json

from collections import defaultdict
from pathlib import Path


import numpy as np


def batch_data(data, batch_size, seed):
    '''
    data is a dict := {'x': [numpy array], 'y': [numpy array]} (on one client)
    returns x, y, which are both numpy array of length: batch_size
    '''
    data_x = data['x']
    data_y = data['y']

    # randomly shuffle data
    np.random.seed(seed)
    rng_state = np.random.get_state()
    np.random.shuffle(data_x)
    np.random.set_state(rng_state)
    np.random.shuffle(data_y)

    # loop through mini-batches
    for i in range(0, len(data_x), batch_size):
        batched_x = data_x[i:i+batch_size]
        batched_y = data_y[i:i+batch_size]
        yield (batched_x, batched_y)


def read_dir(data_dir: Path):
    clients = []
    groups = []
    data = defaultdict(lambda : None)

    files = [f for f in data_dir.iterdir() if f.is_file() and f.suffix == ".json"]
    for f in files:
        with open(f, 'r') as inf:
            cdata = json.load(inf)
        clients.extend(cdata['users'])
        if 'hierarchies' in cdata:
            groups.extend(cdata['hierarchies'])
        data.update(cdata['user_data'])

    clients = list(sorted(data.keys()))
    return clients, groups, data


def read_data(train_data_dir, test_data_dir):
    '''parses data in given train and test data directories

    assumes:
    - the data in the input directories are .json files with 
        keys 'users' and 'user_data'
    - the set of train set users is the same as the set of test set users
    
    Return:
        clients: list of client ids
        groups: list of group ids; empty list if none found
        train_data: dictionary of train data
        test_data: dictionary of test data
    '''
    train_clients, train_groups, train_data = read_dir(train_data_dir)
    test_clients, test_groups, test_data = read_dir(test_data_dir)

    assert train_clients == test_clients
    assert train_groups == test_groups

    return train_clients, train_groups, train_data, test_data



from pathlib import Path
from typing import Union


import torch

from torch import nn, optim
from torch.utils.data import Dataset, default_collate


from rayleaf.tensorarray.tensorarray import TensorArray


class Model(nn.Module):

    def __init__(self, seed: float, lr: float, optimizer=optim.SGD) -> None:
        super(Model, self).__init__()

        self.lr = lr
        self.seed = seed
        self.optimizer = optimizer

        self.flops = 0

        self.collate_fn = default_collate

        self._shapes = None


    @property
    def shapes(self) -> tuple:
        if self._shapes is None:
            self._shapes = self.get_params().shapes

        return self._shapes


    def generate_dataset(self, data: dict, dataset_dir: Path) -> Dataset:
        return None


    def get_params(self) -> TensorArray:
        return TensorArray(self.parameters())
    

    def set_params(self, new_params: Union[list, TensorArray]) -> None:
        with torch.no_grad():
            if isinstance(new_params, TensorArray):
                new_params = new_params.tensors
            for i, param_tensor in enumerate(self.parameters()):
                param_tensor.copy_(new_params[i])



import os

from datetime import datetime
from pathlib import Path


import pandas as pd


import rayleaf.stats as stats
import rayleaf.utils.logging_utils as logging_utils


def teardown(
    save_model: bool,
    output_dir: str,
    dataset: str,
    model,
    server,
    start_time
):
    if save_model:
        # Save server model
        ckpt_path = Path(output_dir, "checkpoints", dataset)
        if not ckpt_path.is_dir():
            os.makedirs(ckpt_path, exist_ok=True)
        save_path = Path(ckpt_path, f"{model}.ckpt")
        save_path = server.save_model(save_path)
        logging_utils.log(f"Model saved in path: {save_path}")

    pd.DataFrame(server.client_flops).to_csv(stats.FLOPS_CSV(stats.STATS_DIR(output_dir)), mode="a+")
    
    end_time = datetime.now()
    logging_utils.log(f"Total Experiment time: {end_time - start_time}")

    logging_utils.shutdown_log()



from pathlib import Path
from typing import Union


import torch

from thop import profile
from torch import Tensor
from torch.utils.data import DataLoader


import rayleaf.entities.constants as constants
import rayleaf.models.utils as model_utils
import rayleaf.stats as stats

from rayleaf.tensorarray.tensorarray import TensorArray


class Client:
    
    def __init__(
            self,
            client_num: int,
            client_id: str,
            dataset_dir: Path,
            train_data: dict,
            eval_data: dict,
            model: type,
            model_settings: dict,
            group: list = None,
            device: str = "cpu"
        ) -> None:

        self.client_num = client_num

        self.device = device

        self.model_cls = model
        self.model_settings = model_settings
        self.model = model(**self.model_settings)

        self.id = client_id
        self.group = group

        self.train_data = self.model.generate_dataset(train_data, dataset_dir)
        self.eval_data = self.model.generate_dataset(eval_data, dataset_dir)

        self._num_train_samples = len(self.train_data) if self.train_data is not None else 0
        self._num_eval_samples = len(self.eval_data) if self.eval_data is not None else 0
        self._num_samples = self.num_train_samples + self.num_eval_samples

        self.metrics = {}
        self.flops = 0
        self.flops_counted = False
        self.delete_model_on_completion = False

        self.init()

    
    def init(self) -> None:
        pass


    def _train(self, server_update: list, num_epochs: int = 1, batch_size: int = 10) -> tuple:
        self.num_epochs = num_epochs
        self.batch_size = batch_size

        self.metrics = {}

        update = self.train(server_update=server_update)

        training_result = {
            constants.CLIENT_ID_KEY: self.id,
            constants.METRICS_KEY: self.metrics,
            constants.UPDATE_KEY: update
        }

        if self.delete_model_on_completion:
            del self.model
        
        return training_result


    def train(self, server_update):
        self.model_params = server_update
        self.train_model()

        return {
            constants.MODEL_PARAMS_KEY: self.model_params,
            constants.NUM_SAMPLES_KEY: self.num_train_samples
        }


    def train_model(self, compute_grads: bool = False):
        if compute_grads:
            old_params = self.model_params

        self.model.to(self.device)

        pin_memory = self.device == "cuda"
        train_dataloader = DataLoader(
            self.train_data, 
            batch_size=self.batch_size, 
            shuffle=False, 
            pin_memory=pin_memory,
            collate_fn=self.model.collate_fn
        )

        self.model.train()
        for _ in range(self.num_epochs):
            self.run_epoch(train_dataloader)

        self.model.to("cpu")

        self.collect_metric(self.flops, constants.FLOPS_KEY)

        if compute_grads:
            return self.model_params - old_params


    def run_epoch(self, dataloader: DataLoader) -> None:

        flops_counted = self.flops_counted

        for X, y in dataloader:
            X, y = X.to(self.device), y.to(self.device)

            if not flops_counted:
                macs, _ = profile(self.model, inputs=(X, ), verbose=False)
                self.flops += macs * 2

            self.run_minibatch(X, y)
        
        self.flops_counted = True


    def run_minibatch(self, X, y):
        probs = self.model.forward(X).squeeze(dim=1)

        self.model.optimizer.zero_grad()
        loss = self.compute_loss(probs, y)
        loss.backward()
        self.model.optimizer.step()


    def compute_loss(self, probs: Tensor, targets: Tensor) -> Tensor:
        return self.model.loss_fn(probs, targets)


    def _eval(self, set_to_use: str ="test", batch_size: int = 10) -> dict:
        assert set_to_use in ["train", "test", "val"]

        if set_to_use == "train":
            data = self.train_data
        elif set_to_use == "test" or set_to_use == "val":
            data = self.eval_data

        self.model.to(self.device)

        eval_metrics = self.eval_model(data, batch_size)

        self.model.to("cpu")

        eval_metrics[stats.CLIENT_ID_KEY] = self.id

        if self.delete_model_on_completion:
            del self.model
        
        return eval_metrics


    @torch.no_grad()
    def eval_model(self, data, batch_size: int = 10) -> dict:
        pin_memory = self.device == "cuda"
        test_dataloader = DataLoader(
            data, 
            batch_size=batch_size, 
            shuffle=False, 
            pin_memory=pin_memory,
            collate_fn=self.model.collate_fn
        )

        size = len(data)
        num_batches = len(test_dataloader)

        self.model.eval()
        test_loss, correct = 0, 0

        for X, y in test_dataloader:
            X, y = X.to(self.device), y.to(self.device)
            
            probs = self.model.forward(X).squeeze(dim=1)
            test_loss += self.model.loss_fn(probs, y).item()

            preds = model_utils.get_predicted_labels(probs)
            correct += model_utils.number_of_correct(preds, y)

        test_loss /= num_batches

        return {
            stats.NUM_CORRECT_KEY: correct,
            stats.NUM_SAMPLES_KEY: size,
            stats.LOSS_KEY: test_loss
        }


    @property
    def model_params(self) -> TensorArray:
        return self.model.get_params()


    @model_params.setter
    def model_params(self, params: Union[list, TensorArray]) -> None:
        if not hasattr(self, "model"):
            self.model = self.model_cls(**self.model_settings)
        self.model.set_params(params)


    @property
    def num_train_samples(self) -> int:
        return self._num_train_samples


    @property
    def num_eval_samples(self) -> int:
        return self._num_eval_samples

    
    @property
    def num_samples(self) -> int:
        return self._num_samples


    def collect_metric(self, metric, metric_name):
        self.metrics[metric_name] = metric



import os

from pathlib import Path
from typing import Union


from.graphing import line_plot


"""
Raw client statistic keys
"""
CLIENT_ID_KEY = "client_id"
ROUND_NUMBER_KEY = "round_number"
NUM_CORRECT_KEY = "num_correct"
NUM_SAMPLES_KEY = "num_samples"
LOSS_KEY = "loss"


"""
Aggregate statistics keys
"""
AVERAGE_KEY = "average"
MEDIAN_KEY = "median"
def PERCENTILE_KEY(percentile: int) -> str:
    last_digit = percentile % 10
    if last_digit == 1:
        suffix = "st"
    elif last_digit == 2:
        suffix = "nd"
    elif last_digit == 3:
        suffix = "rd"
    else:
        suffix = "th"

    return f"{percentile}{suffix} percentile"

    
"""
Path constants
"""
def STATS_DIR(output_dir: Union[str, Path]) -> Path:
    stats_dir = Path(output_dir, "stats")
    if not stats_dir.is_dir():
        os.makedirs(stats_dir, exist_ok=True)
    
    return stats_dir


def STATS_CSV(stats_dir: Union[str, Path], stat_set: str, aggregate: bool) -> Path:
    csv_name = f"{stat_set}_stats.csv"
    if aggregate:
        csv_name = f"agg_{csv_name}"
    
    return Path(stats_dir, csv_name)


def FLOPS_CSV(stats_dir: Union[str, Path]) -> Path:
    csv_name = "client_flop_counts.csv"

    return Path(stats_dir, csv_name)


def GRAPHS_DIR(output_dir: Union[str, Path]) -> Path:
    stats_dir = Path(output_dir, "graphs")
    if not stats_dir.is_dir():
        os.makedirs(stats_dir, exist_ok=True)
    
    return stats_dir


def METRICS_DIR(output_dir: Union[str, Path]) -> Path:
    metrics_dir = Path(output_dir, "metrics")
    if not metrics_dir.is_dir():
        os.makedirs(metrics_dir, exist_ok=True)
    
    return metrics_dir



from pathlib import Path


import numpy as np
import torch

from PIL import Image
from torch import nn, Tensor
import torch.nn.functional as F
from torch.utils.data import TensorDataset


from rayleaf.models.model import Model


IMAGE_SIZE = 84


class ClientModel(Model):
    def __init__(self, seed, lr, num_classes):
        super(ClientModel, self).__init__(seed, lr)

        self.num_classes = num_classes

        self.conv1 = nn.Conv2d(in_channels=3, out_channels=32, kernel_size=3, padding="same")
        self.bn1 = nn.BatchNorm2d(num_features=32)
        self.pool1 = nn.MaxPool2d(kernel_size=(2, 2), stride=2)

        self.conv2 = nn.Conv2d(in_channels=32, out_channels=32, kernel_size=3, padding="same")
        self.bn2 = nn.BatchNorm2d(num_features=32)
        self.pool2 = nn.MaxPool2d(kernel_size=(2, 2), stride=2)

        self.conv3 = nn.Conv2d(in_channels=32, out_channels=32, kernel_size=3, padding="same")
        self.bn3 = nn.BatchNorm2d(num_features=32)
        self.pool3 = nn.MaxPool2d(kernel_size=(2, 2), stride=2)

        self.conv4 = nn.Conv2d(in_channels=32, out_channels=32, kernel_size=3, padding="same")
        self.bn4 = nn.BatchNorm2d(num_features=32)
        self.pool4 = nn.MaxPool2d(kernel_size=(2, 2), stride=2)

        self.fc1 = nn.Linear(32 * 5 * 5, self.num_classes)

        self.loss_fn = nn.CrossEntropyLoss()
        self.optimizer = self.optimizer(self.parameters(), lr=self.lr)


    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.pool1(x)
        x = F.relu(x)

        x = self.conv2(x)
        x = self.bn2(x)
        x = self.pool2(x)
        x = F.relu(x)

        x = self.conv3(x)
        x = self.bn3(x)
        x = self.pool3(x)
        x = F.relu(x)

        x = self.conv4(x)
        x = self.bn4(x)
        x = self.pool4(x)
        x = F.relu(x)

        x = torch.flatten(x, start_dim=1)
        logits = self.fc1(x)

        return logits
        

    def generate_dataset(self, data: dict, dataset_dir: Path):
        img_dir = Path(dataset_dir, "data", "raw", "img_align_celeba")
        assert img_dir.is_dir()

        X = []
        for file_name in data["x"]:
            X.append(self._load_image(Path(img_dir, file_name)))

        X = torch.stack(X)
        y = Tensor(data["y"]).long()

        return TensorDataset(X, y)


    def _load_image(self, img_path: Path):
        img = Image.open(img_path)
        img = img.resize((IMAGE_SIZE, IMAGE_SIZE)).convert('RGB')

        img = Tensor(np.array(img)).permute((2, 0, 1))

        return img



import torch


import rayleaf

from rayleaf.models.femnist.cnn import ClientModel as CNN


def test_simple():
    t1 = torch.zeros(1, 2, 3, 4)
    t2 = torch.zeros(5)

    ta = rayleaf.TensorArray([t1, t2])
    assert len(ta) == 2


def test_model():
    model = CNN(0, 0.05, 62)
    
    ta = rayleaf.TensorArray(model)
    assert len(ta) == len(tuple(model.parameters()))



import numpy as np
import ray
import torch

from tqdm import tqdm


import rayleaf.entities.constants as constants

from rayleaf.tensorarray.tensorarray import TensorArray


class Server:
    
    def __init__(self, model_params: list, client_clusters: list) -> None:
        self.model_params = TensorArray(model_params)
        self.layer_shapes = self.model_params.shapes

        self.client_clusters = client_clusters
        self.num_client_clusters = len(client_clusters) 

        self.updates = []
        self.selected_clients = [[] for _ in range(self.num_client_clusters)]

        self.clients_profiled = set()
        self.client_flops = []

        self.init()


    def init(self) -> None:
        pass


    def select_clients(self, my_round: int, possible_clients: list, num_clients: int = 20) -> None:
        selected_client_nums = np.random.choice(possible_clients, num_clients, replace=False)
        self.selected_clients = [[] for _ in range(self.num_client_clusters)]

        for client_num in selected_client_nums:
            self.selected_clients[client_num % self.num_client_clusters].append(client_num)
        
        return selected_client_nums


    def server_update(self):
        return self.model_params


    def train_clients(self, num_epochs: int = 1, batch_size: int = 10) -> None:
        training_futures = []
        for client_cluster_idx, cluster_clients in enumerate(self.selected_clients):
            if len(cluster_clients) > 0:
                training_future = self.client_clusters[client_cluster_idx].train_clients.remote(
                    server_update=self.server_update(),
                    selected_clients=cluster_clients,
                    num_epochs=num_epochs,
                    batch_size=batch_size
                )
                training_futures.append(training_future)
        
        num_futures = len(training_futures)
        with tqdm(total=num_futures, leave=False, desc="Training clients") as pbar:
            while len(training_futures) > 0:
                complete, incomplete = ray.wait(training_futures)

                for result in ray.get(complete):
                    self.updates.extend(result)
                    pbar.update(1)

                training_futures = incomplete
        
        return self.updates


    def update_model(self, client_updates):
        num_samples = 0
        average_params = 0

        for update in client_updates:
            average_params += update[constants.MODEL_PARAMS_KEY] * update[constants.NUM_SAMPLES_KEY]
            num_samples += update[constants.NUM_SAMPLES_KEY]
        
        average_params /= num_samples

        return average_params


    @torch.no_grad()
    def _update_model(self) -> None:
        client_updates = []
        for update in self.updates:
            client_updates.append(update[constants.UPDATE_KEY])

        self.model_params = self.update_model(client_updates=client_updates)
        self.updates.clear()


    def eval_model(self, eval_all_clients: bool = True, set_to_use: str = "test", batch_size: int = 10) -> dict:
        eval_futures = []
        if eval_all_clients:
            for client_cluster in self.client_clusters:
                eval_future = client_cluster.eval_model.remote(
                    model_params=self.model_params,
                    set_to_use=set_to_use,
                    batch_size=batch_size 
                )
                eval_futures.append(eval_future)
        else:
            for client_cluster_idx, cluster_clients in enumerate(self.selected_clients):
                if len(cluster_clients) > 0:
                    eval_future = self.client_clusters[client_cluster_idx].eval_model.remote(
                        model_params=self.model_params,
                        set_to_use=set_to_use,
                        selected_clients=cluster_clients,
                        batch_size=batch_size
                    )
                    eval_futures.append(eval_future)

        stats = []

        num_futures = len(eval_futures)
        with tqdm(total=num_futures, leave=False, desc=f"Evaluating model on {set_to_use} set") as pbar:
            while len(eval_futures) > 0:
                complete, incomplete = ray.wait(eval_futures)

                for cluster_metrics in ray.get(complete):
                    stats.extend(cluster_metrics)
                    pbar.update(1)

                eval_futures = incomplete
        
        return stats


    def get_clients_info(self) -> tuple:
        info_futures = []
        for client_cluster in self.client_clusters:
            info_future = client_cluster.get_clients_info.remote()
            info_futures.append(info_future)

        ids = []
        groups = {}
        num_samples = {}
        while len(info_futures) > 0:
            complete, incomplete = ray.wait(info_futures)

            for future_ids, future_groups, future_num_samples in ray.get(complete):
                ids.extend(future_ids)
                groups.update(future_groups)
                num_samples.update(future_num_samples)

            info_futures = incomplete
        
        return ids, groups, num_samples


    def save_model(self, path: str) -> None:
        torch.save({"model_params": self.model_params}, path)


    @property
    def model_params(self) -> TensorArray:
        return self._model_params

    
    @model_params.setter
    def model_params(self, params: TensorArray) -> None:
        self._model_params = params



from typing import Union


from numpy import ndarray
from scipy.linalg import block_diag
from torch import stack, Tensor


def _tensor_to_block_diag(tensor: Union[ndarray, Tensor]) -> Tensor:
    if len(tensor.shape) <= 2:
        return tensor
    
    mats = []
    for i in range(tensor.shape[0]):
        mats.append(_tensor_to_block_diag(tensor[i]))
    
    return Tensor(block_diag(*mats)).requires_grad_(False)


def _block_diag_to_tensor(mat: Union[ndarray, Tensor], shape: tuple, start_inds: list = [0, 0]) -> Tensor:
    row = start_inds[0]
    col = start_inds[1]

    if len(shape) == 1:
        res = Tensor(mat[row][col:col + shape[0]]).requires_grad_(False)
        start_inds[0] += 1
        start_inds[1] += shape[0]
        return res
    if len(shape) == 2:
        res = Tensor(mat[row:row + shape[0], col:col + shape[1]]).requires_grad_(False)
        start_inds[0] += shape[0]
        start_inds[1] += shape[1]
        return res

    tensors = []

    for _ in range(shape[0]):
        tensors.append(_block_diag_to_tensor(mat, shape[1:], start_inds))
    
    return stack(tensors).requires_grad_(False)




import torch
import numpy as np

from torch import Tensor


import rayleaf

from rayleaf.models.femnist.cnn import ClientModel as CNN
from rayleaf.models.speech_commands.m5 import ClientModel as M5


def test_simple_utils():
    shape = (3, 2, 3)

    arr = torch.arange(np.prod(shape)).reshape(shape)

    blocked = rayleaf.tensorarray.utils._tensor_to_block_diag(arr)
    expected = Tensor([
        [0, 1, 2, 0, 0, 0, 0, 0, 0],
        [3, 4, 5, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 6, 7, 8, 0, 0, 0],
        [0, 0, 0, 9, 10, 11, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 12, 13, 14],
        [0, 0, 0, 0, 0, 0, 15, 16, 17]
    ])
    assert torch.all(blocked == expected)

    reconstructed = rayleaf.tensorarray.utils._block_diag_to_tensor(blocked, shape)
    assert torch.all(reconstructed == arr)


def test_simple_ta():
    shapes = [(2, 2, 3), (3, 2), (2, 1, 2)]

    tensors = [torch.arange(np.prod(shape)).reshape(shape) for shape in shapes]
    ta = rayleaf.TensorArray(tensors)

    blocked = ta.as_block_diag()
    expected = Tensor([
        [0, 1, 2, 0, 0,  0,  0, 0, 0, 0, 0, 0],
        [3, 4, 5, 0, 0,  0,  0, 0, 0, 0, 0, 0],
        [0, 0, 0, 6, 7,  8,  0, 0, 0, 0, 0, 0],
        [0, 0, 0, 9, 10, 11, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0,  0,  0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0,  0,  2, 3, 0, 0, 0, 0],
        [0, 0, 0, 0, 0,  0,  4, 5, 0, 0, 0, 0],
        [0, 0, 0, 0, 0,  0,  0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 2, 3],
    ])
    assert torch.all(blocked == expected)

    reconstructed = rayleaf.TensorArray.from_block_diag(blocked, shapes)
    assert reconstructed == ta


def test_model_femnist():
    torch.manual_seed(0)
    model = CNN(0, 0.05, 62)
    ta = rayleaf.TensorArray(model)

    print(model.shapes)

    blocked = ta.as_block_diag()
    print("femnist", blocked.shape)
    reconstructed = rayleaf.TensorArray.from_block_diag(blocked, model.shapes)
    assert reconstructed == ta


def test_model_speech_commands():
    torch.manual_seed(0)
    model = M5(0, 0.05, 35)
    ta = rayleaf.TensorArray(model)

    print(model.shapes)

    blocked = ta.as_block_diag()
    print("speech_commands", blocked.shape)
    reconstructed = rayleaf.TensorArray.from_block_diag(blocked, model.shapes)
    assert reconstructed == ta



"""dict: Specifies execution parameters (tot_num_rounds, eval_every_num_rounds, clients_per_round)"""

MODEL_SETTINGS = {
    "sent140.bag_dnn": {
        "num_classes": 2
    },
    "sent140.stacked_lstm": {
        "seq_len": 25,
        "num_classes": 2,
        "num_hidden": 100
    },
    "sent140.bag_log_reg": {
        "num_classes": 2
    },
    "femnist.cnn": {
        "num_classes": 62
    },
    "shakespeare.stacked_lstm": {
        "seq_len": 80,
        "num_classes": 80,
        "num_hidden": 256
    },
    "celeba.cnn": {
        "num_classes": 2
    },
    "synthetic.log_reg": {
        "num_classes": 5,
        "input_dim": 60
    },
    "reddit.stacked_lstm": {
        "seq_len": 10,
        "num_hidden": 256,
        "num_layers": 2
    },
    "speech_commands.m5": {
        "num_classes": 35
    }
}



from datetime import datetime
from pathlib import Path


import rayleaf
from rayleaf.entities import Server, Client


def fedavg(
    dataset: str,
    output_dir: str,
    num_rounds: int = 100,
    eval_every: int = 10,
    num_clients: int = 200,
    clients_per_round: int = 40,
    client_lr: float = 0.05,
    batch_size: int = 64,
    seed: int = 0,
    num_epochs: int = 10,
    gpus_per_client_cluster: float = 1,
    num_client_clusters: int = 8,
    save_model: bool = False,
    notes: str = None
):
    curr_time = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")

    if dataset == "femnist":
        model = "cnn"
    elif dataset == "speech_commands":
        model = "m5"
    elif dataset == "shakespeare":
        model = "stacked_lstm"

    rayleaf.run_experiment(
        dataset = dataset,
        dataset_dir = f"/home/huberkeley/rayleaf/data/{dataset}/",
        output_dir= Path(output_dir, dataset, "fedavg", f"{num_clients}clients-{clients_per_round}cpr-{client_lr}lr-{num_epochs}epochs-{num_rounds}rounds-{curr_time}"),
        model = model,
        num_rounds = num_rounds,
        eval_every = eval_every,
        ServerType=Server,
        client_types=[(Client, num_clients)],
        clients_per_round = clients_per_round,
        client_lr = client_lr,
        batch_size = batch_size,
        seed = seed,
        use_val_set = False,
        num_epochs = num_epochs,
        gpus_per_client_cluster = gpus_per_client_cluster,
        num_client_clusters = num_client_clusters,
        save_model = save_model,
        notes = notes
    )



from pathlib import Path


import numpy as np
import pandas as pd


import rayleaf.utils as utils
import rayleaf.utils.logging_utils as logging_utils
from rayleaf.entities.server import Server
import rayleaf.stats as stats


def eval_server(
    server: Server,
    num_round: int,
    eval_set: bool,
    output_dir: str,
    batch_size: int
):
    output_dir = Path(output_dir)
    stats_dir = stats.STATS_DIR(output_dir=output_dir)

    train_stats, eval_stats = _eval_clients(server, num_round, eval_set=eval_set, batch_size=batch_size)

    _append_to_csv(train_stats, stats_dir, "train", aggregate=False)
    _append_to_csv(eval_stats, stats_dir, eval_set, aggregate=False)

    agg_train_stats, agg_eval_stats = _compute_stats(train_stats), _compute_stats(eval_stats)
    _print_aggregate_stats(agg_train_stats, header=f"training accuracy: round {num_round}".title())
    _print_aggregate_stats(agg_eval_stats, header=f"{eval_set} accuracy: round {num_round}".title())
    agg_train_stats[stats.ROUND_NUMBER_KEY], agg_eval_stats[stats.ROUND_NUMBER_KEY] = num_round, num_round

    _append_to_csv(agg_train_stats, stats_dir, "train", aggregate=True)
    _append_to_csv(agg_eval_stats, stats_dir, eval_set, aggregate=True)


def _append_to_csv(
    stats_df: pd.DataFrame,
    stats_dir: Path,
    dataset: str,
    aggregate: bool
):
    csv_file = stats.STATS_CSV(stats_dir, dataset, aggregate=aggregate)
    append_header = not csv_file.is_file()
    stats_df.to_csv(csv_file, mode="a+", index=False, header=append_header)


def _eval_clients(
    server: Server,
    num_round: int,
    eval_set: str,
    batch_size: int
):
    train_stats = server.eval_model(set_to_use="train", batch_size=batch_size)
    eval_stats = server.eval_model(set_to_use=eval_set, batch_size=batch_size)

    train_stats, eval_stats = pd.DataFrame(train_stats), pd.DataFrame(eval_stats)
    train_stats[stats.ROUND_NUMBER_KEY], eval_stats[stats.ROUND_NUMBER_KEY] = num_round, num_round

    return train_stats, eval_stats


def _compute_stats(
    stats_df: pd.DataFrame
):
    num_correct = stats_df[stats.NUM_CORRECT_KEY].values
    num_samples = stats_df[stats.NUM_SAMPLES_KEY].values

    accuracy = num_correct / num_samples

    avg = np.average(accuracy, weights=num_samples)
    med = np.median(accuracy)
    percentiles = {
        10: np.nanpercentile(accuracy, 10),
        90: np.nanpercentile(accuracy, 90)
    }

    aggregate_stats = {
        stats.AVERAGE_KEY: avg,
        stats.MEDIAN_KEY: med
    }
    for percentile, value in percentiles.items():
        aggregate_stats[stats.PERCENTILE_KEY(percentile)] = value

    return pd.DataFrame([aggregate_stats])


def _print_aggregate_stats(stats: pd.DataFrame, header: str):
    logging_utils.log(utils.EVAL_STR.format(header))

    logging_utils.log_df(stats)



from fileinput import filename
from pathlib import Path
from typing import Union


import numpy as np

from matplotlib import pyplot as plt


def line_plot(
    output_dir: Union[Path, str],
    title: str,
    x: Union[np.ndarray, list],
    xlabel: str,
    ylabel: str,
    marker: str,
    **kwargs
):
    file_name = title.lower().replace(" ", "_")
    graph_path = Path(output_dir, f"{file_name}.png")

    fig = plt.figure()

    for k, y in kwargs.items():
        plt.plot(x, y, label=k, marker=marker)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.legend()
    plt.savefig(graph_path)
    plt.close(fig)



import torch


import rayleaf


def test_modify_tensor_after_creating_ta_simple():
    t = torch.ones(4, 4)
    ta1 = rayleaf.TensorArray([t])
    t[0][0] = 0
    ta2 = rayleaf.TensorArray([t])

    assert ta1 != ta2



import json
import os

from pathlib import Path
from tqdm import tqdm


import torch

from torchaudio.datasets import SPEECHCOMMANDS


MAX_NUMBER_OF_CLIENTS = 2000


class SpeakerSC(SPEECHCOMMANDS):
    def __init__(self, root: Path, speaker: str):
        super().__init__(root=root, download=True)
        
        self._walker = [w for w in self._walker if speaker in w]


def federate_dataset(root, seed, num_clients=-1, train_split=0.8, iid=False):
    assert -1 <= num_clients <= MAX_NUMBER_OF_CLIENTS, f"num_clients must be an integer between -1 and {MAX_NUMBER_OF_CLIENTS} for speech_commands dataset"

    if num_clients == -1:
        num_clients = MAX_NUMBER_OF_CLIENTS

    root = Path(root)
    if not root.is_dir():
        os.makedirs(root, exist_ok=True)

    if not iid:
        return federate_dataset_realistic(root, seed, num_clients, train_split)
    else:
        return federate_dataset_iid(root, seed, num_clients, train_split)


def federate_dataset_realistic(root: Path, seed: float, num_clients: int, train_split: float):
    speaker_data_size_json = Path(root, "speaker_data_size.json")
        
    if not speaker_data_size_json.is_file():
        data = SPEECHCOMMANDS(
            root=root,
            download=True
        )
        
        speakers = set()

        for d in tqdm(data, desc="Collecting speaker IDs", leave=False):
            speakers.add(d[3])
        
        speakers = list(speakers)
        
        speaker_data_size = {}
        speaker_dataset = {}
        
        for speaker in tqdm(speakers, desc="Counting speaker dataset sizes", leave=False):
            data = SpeakerSC(
                root=root,
                speaker=speaker
            )
            speaker_data_size[speaker] = len(data)
            speaker_dataset[speaker] = data
        
        with open(speaker_data_size_json, mode="w+") as fp:
            json.dump(speaker_data_size, fp)
        
        speakers.sort(key=lambda speaker: speaker_data_size[speaker], reverse=True)
        
        datasets = {}
        for speaker in tqdm(speakers[:num_clients], "Federating dataset", leave=False):
            datasets[speaker] = speaker_dataset[speaker]
    else:
        with open(speaker_data_size_json, "r") as fp:
            print("Loading speaker to dataset count json...")
            speaker_data_size = json.load(fp)
        
        speakers = sorted(list(speaker_data_size.keys()), key=lambda speaker: speaker_data_size[speaker], reverse=True)
            
        datasets = {}
        for speaker in tqdm(speakers[:num_clients], desc="Federating dataset", leave=False):
            datasets[speaker] = SpeakerSC(
                                    root=root,
                                    speaker=speaker
                                )
        
    users = list(datasets.keys())
    groups = None
    train_data, test_data = {}, {}

    for user in tqdm(users, "Generating training and testing splits", leave=False):
        user_dataset = datasets[user]

        total_size = len(user_dataset)
        train_size = int(total_size * train_split)
        test_size = total_size - train_size

        train_data[user], test_data[user] = torch.utils.data.random_split(user_dataset, [train_size, test_size], generator=torch.Generator().manual_seed(seed))
    
    return users, groups, train_data, test_data


def federate_dataset_iid(root: Path, seed: float, num_clients: int, train_split: float):
    data = SPEECHCOMMANDS(
        root=root,
        download=True
    )
    
    users = []
    groups = None
    train_data, test_data = {}, {}

    n = len(data)
    data_per_client = len(data) // num_clients

    split = [data_per_client + 1  if i < n % num_clients else data_per_client for i in range(num_clients)]

    federated = torch.utils.data.random_split(data, split, generator=torch.Generator().manual_seed(seed))

    for client in range(num_clients):
        user = str(client)

        users.append(user)

        train_size = int(split[client]  * train_split)
        test_size = split[client] - train_size
        train_data[user], test_data[user] = torch.utils.data.random_split(federated[client], [train_size, test_size], generator=torch.Generator().manual_seed(seed))

    return users, groups, train_data, test_data



"""Utils for language models."""

import re
import json


import numpy as np

# ------------------------
# utils for shakespeare dataset

ALL_LETTERS = "\n !\"&'(),-.0123456789:;>?ABCDEFGHIJKLMNOPQRSTUVWXYZ[]abcdefghijklmnopqrstuvwxyz}"
NUM_LETTERS = len(ALL_LETTERS)


def _one_hot(index, size):
    '''returns one-hot vector with given size and value 1 at given index
    '''
    vec = [0 for _ in range(size)]
    vec[int(index)] = 1
    return vec

def letter_to_index(letter):
    '''returns index of given letter
    '''
    return ALL_LETTERS.find(letter)

def letter_to_vec(letter):
    '''returns one-hot representation of given letter
    '''
    index = ALL_LETTERS.find(letter)
    return _one_hot(index, NUM_LETTERS)


def word_to_indices(word):
    '''returns a list of character indices

    Args:
        word: string
    
    Return:
        indices: int list with length len(word)
    '''
    indices = []
    for c in word:
        indices.append(ALL_LETTERS.find(c))
    return indices


# ------------------------
# utils for sent140 dataset


def split_line(line):
    '''split given line/phrase into list of words

    Args:
        line: string representing phrase to be split
    
    Return:
        list of strings, with each string representing a word
    '''
    return re.findall(r"[\w']+|[.,!?;]", line)


def _word_to_index(word, indd):
    '''returns index of given word based on given lookup dictionary

    returns the length of the lookup dictionary if word not found

    Args:
        word: string
        indd: dictionary with string words as keys and int indices as values
    '''
    if word in indd:
        return indd[word]
    else:
        return len(indd)


def line_to_indices(line, word2id, max_words=25):
    '''converts given phrase into list of word indices
    
    if the phrase has more than max_words words, returns a list containing
    indices of the first max_words words
    if the phrase has less than max_words words, repeatedly appends integer 
    representing unknown index to returned list until the list's length is 
    max_words

    Args:
        line: string representing phrase/sequence of words
        word2id: dictionary with string words as keys and int indices as values
        max_words: maximum number of word indices in returned list

    Return:
        indl: list of word indices, one index for each word in phrase
    '''
    unk_id = len(word2id)
    line_list = split_line(line) # split phrase in words
    indl = [word2id[w] if w in word2id else unk_id for w in line_list[:max_words]]
    indl += [unk_id]*(max_words-len(indl))
    return indl


def bag_of_words(line, vocab):
    '''returns bag of words representation of given phrase using given vocab

    Args:
        line: string representing phrase to be parsed
        vocab: dictionary with words as keys and indices as values

    Return:
        integer list
    '''
    bag = [0]*len(vocab)
    words = split_line(line)
    for w in words:
        if w in vocab:
            bag[vocab[w]] += 1
    return bag


def get_word_emb_arr(path):
    with open(path, 'r') as inf:
        embs = json.load(inf)
    vocab = embs['vocab']
    word_emb_arr = np.array(embs['emba'])
    indd = {}
    for i in range(len(vocab)):
        indd[vocab[i]] = i
    vocab = {w: i for i, w in enumerate(embs['vocab'])}
    return word_emb_arr, indd, vocab


def val_to_vec(size, val):
    """Converts target into one-hot.

    Args:
        size: Size of vector.
        val: Integer in range [0, size].
    Returns:
         vec: one-hot vector with a 1 in the val element.
    """
    assert 0 <= val < size
    vec = [0 for _ in range(size)]
    vec[int(val)] = 1
    return vec



import numpy as np
import torch


import rayleaf

from rayleaf.models.femnist.cnn import ClientModel as CNN


def test_simple_eq():
    rng = np.random.RandomState(0)
    arr1 = rng.rand(3, 1, 4, 1)
    arr2 = rng.rand(5, 9)
    arr3 = rng.rand(2, 6, 5)
    arr4 = rng.rand(3, 5, 8)
    arr = [arr1, arr2, arr3, arr4]

    ta1 = rayleaf.TensorArray(arr)

    rng = np.random.RandomState(0)
    arr1 = rng.rand(3, 1, 4, 1)
    arr2 = rng.rand(5, 9)
    arr3 = rng.rand(2, 6, 5)
    arr4 = rng.rand(3, 5, 8)
    arr = [arr1, arr2, arr3, arr4]
    ta2 = rayleaf.TensorArray(arr)

    assert ta1 == ta2


def test_model_eq():
    torch.manual_seed(0)
    model = CNN(0, 0.05, 62)
    ta1 = rayleaf.TensorArray(model)

    torch.manual_seed(0)
    model = CNN(0, 0.05, 62)
    ta2 = rayleaf.TensorArray(model)

    assert ta1 == ta2


def test_type_mismatch():
    torch.manual_seed(0)
    model = CNN(0, 0.05, 62)
    ta1 = rayleaf.TensorArray(model)

    ta2 = 0

    assert not ta1 == ta2


def test_len_mismatch():
    torch.manual_seed(0)
    model = CNN(0, 0.05, 62)
    ta1 = rayleaf.TensorArray(model)

    ta2 = rayleaf.TensorArray([[0]])

    assert not ta1 == ta2


def test_shape_mismatch():
    rng = np.random.RandomState(0)
    arr1 = rng.rand(3, 1, 4, 1)
    arr2 = rng.rand(5, 9)
    arr3 = rng.rand(2, 6, 5)
    arr4 = rng.rand(3, 5, 8)
    arr = [arr1, arr2, arr3, arr4]

    ta1 = rayleaf.TensorArray(arr)

    rng = np.random.RandomState(0)
    arr1 = rng.rand(3, 1, 4, 1)
    arr2 = rng.rand(5, 9)
    arr3 = rng.rand(2, 6, 5)
    arr4 = rng.rand(3, 5, 9)
    arr = [arr1, arr2, arr3, arr4]
    ta2 = rayleaf.TensorArray(arr)

    assert not ta1 == ta2



from .setup_stage import initialize_resources, log_experiment_details, create_entities
from .train_stage import train
from .eval_stage import eval_server
from .graph_stage import graph
from .teardown_stage import teardown



from pathlib import Path


import pandas as pd


import rayleaf.stats as stats
import rayleaf.utils as utils
import rayleaf.utils.logging_utils as logging_utils

from rayleaf.entities.server import constants, Server
from rayleaf.utils.client_count_utils import client_counts_string, count_selected_clients, online


def train(
    server: Server,
    clients: list,
    clients_per_round: int,
    output_dir: str,
    num_epochs: int,
    batch_size: int,
    curr_round: int,
    num_rounds: int,
    log_progress: bool = True
):
    selected_clients = server.select_clients(curr_round, online(clients), num_clients=clients_per_round)
    if log_progress:
        selected_client_counts = count_selected_clients(selected_clients, clients)
        logging_utils.log(utils.ROUND_STR.format(f"Round {curr_round} of {num_rounds}: Training {clients_per_round} clients: {client_counts_string(selected_client_counts)}"))

    updates = server.train_clients(num_epochs=num_epochs, batch_size=batch_size)

    metrics_dir = stats.METRICS_DIR(output_dir)
    for update in updates:
        cid = update[constants.CLIENT_ID_KEY]
        for metric_name in update[constants.METRICS_KEY]:
            metric = pd.DataFrame([
                {
                    constants.CLIENT_ID_KEY: cid,
                    constants.ROUND_KEY: curr_round,
                    metric_name: update[constants.METRICS_KEY][metric_name]
                }
            ])
            _append_metric(metric, metric_name, metrics_dir)

    server._update_model()


def _append_metric(metric_df: pd.DataFrame, metric_name: str, metrics_dir: Path):
    csv_file = Path(metrics_dir, f"{metric_name}.csv")
    append_header = not csv_file.is_file()
    metric_df.to_csv(csv_file, mode="a+", index=False, header=append_header)



import numpy as np


import rayleaf


def test_simple_array_ufunc():
    rng = np.random.RandomState(0)
    arr1 = rng.rand(3, 1, 4, 1)
    arr2 = rng.rand(5, 9)
    arr3 = rng.rand(2, 6, 5)
    arr4 = rng.rand(3, 5, 8)
    arr = [arr1, arr2, arr3, arr4]

    ta1 = rayleaf.TensorArray(arr)

    assert ta1 == 1 * ta1


def test_simple_add():
    shape1 = (3, 1, 4, 1)
    shape2 = (5, 9)
    shape3 = (2, 6, 5)
    shape4 = (3, 5, 8)
    shapes = [shape1, shape2, shape3, shape4]

    ta1 = rayleaf.TensorArray([np.ones(shape) for shape in shapes])
    ta2 = rayleaf.TensorArray([2 * np.ones(shape) for shape in shapes])
    ta3 = rayleaf.TensorArray([3 * np.ones(shape) for shape in shapes])

    assert ta1 + ta2 == ta3
    assert ta1 + 2 == ta3


def test_simple_sub():
    shape1 = (3, 1, 4, 1)
    shape2 = (5, 9)
    shape3 = (2, 6, 5)
    shape4 = (3, 5, 8)
    shapes = [shape1, shape2, shape3, shape4]

    ta1 = rayleaf.TensorArray([np.ones(shape) for shape in shapes])
    ta2 = rayleaf.TensorArray([2 * np.ones(shape) for shape in shapes])
    ta3 = rayleaf.TensorArray([-1 * np.ones(shape) for shape in shapes])

    assert ta1 - ta2 == ta3
    assert ta1 - 2 == ta3


def test_simple_mul():
    shape1 = (3, 1, 4, 1)
    shape2 = (5, 9)
    shape3 = (2, 6, 5)
    shape4 = (3, 5, 8)
    shapes = [shape1, shape2, shape3, shape4]

    ta1 = rayleaf.TensorArray([2 * np.ones(shape) for shape in shapes])
    ta2 = rayleaf.TensorArray([-3 * np.ones(shape) for shape in shapes])
    ta3 = rayleaf.TensorArray([-6 * np.ones(shape) for shape in shapes])

    assert ta1 * ta2 == ta3
    assert ta1 * -3 == ta3


def test_simple_pow():
    shape1 = (3, 1, 4, 1)
    shape2 = (5, 9)
    shape3 = (2, 6, 5)
    shape4 = (3, 5, 8)
    shapes = [shape1, shape2, shape3, shape4]

    ta1 = rayleaf.TensorArray([2 * np.ones(shape) for shape in shapes])
    ta2 = rayleaf.TensorArray([3 * np.ones(shape) for shape in shapes])
    ta3 = rayleaf.TensorArray([8 * np.ones(shape) for shape in shapes])

    assert ta1 ** ta2 == ta3
    assert ta1 ** 3 == ta3


def test_simple_truediv():
    shape1 = (3, 1, 4, 1)
    shape2 = (5, 9)
    shape3 = (2, 6, 5)
    shape4 = (3, 5, 8)
    shapes = [shape1, shape2, shape3, shape4]

    ta1 = rayleaf.TensorArray([6 * np.ones(shape) for shape in shapes])
    ta2 = rayleaf.TensorArray([3 * np.ones(shape) for shape in shapes])
    ta3 = rayleaf.TensorArray([2 * np.ones(shape) for shape in shapes])

    assert ta1 / ta2 == ta3
    assert ta1 / 3 == ta3


def test_simple_floordiv():
    shape1 = (3, 1, 4, 1)
    shape2 = (5, 9)
    shape3 = (2, 6, 5)
    shape4 = (3, 5, 8)
    shapes = [shape1, shape2, shape3, shape4]

    ta1 = rayleaf.TensorArray([8 * np.ones(shape) for shape in shapes])
    ta2 = rayleaf.TensorArray([3 * np.ones(shape) for shape in shapes])
    ta3 = rayleaf.TensorArray([2 * np.ones(shape) for shape in shapes])

    assert ta1 // ta2 == ta3
    assert ta1 // 3 == ta3


def test_simple_modulo():
    shape1 = (3, 1, 4, 1)
    shape2 = (5, 9)
    shape3 = (2, 6, 5)
    shape4 = (3, 5, 8)
    shapes = [shape1, shape2, shape3, shape4]

    ta1 = rayleaf.TensorArray([7 * np.ones(shape) for shape in shapes])
    ta2 = rayleaf.TensorArray([3 * np.ones(shape) for shape in shapes])
    ta3 = rayleaf.TensorArray([1 * np.ones(shape) for shape in shapes])

    assert ta1 % ta2 == ta3
    assert ta1 % 3 == ta3



"""Bag-of-words logistic regression."""

import numpy as np
import os
import sys
import tensorflow as tf

from rayleaf.models.model import Model
from utils.language_utils import bag_of_words, get_word_emb_arr, val_to_vec

VOCAB_DIR = 'sent140/embs.json'


class ClientModel(Model):

    def __init__(self, lr, num_classes, input_dim=None):
        self.num_classes = num_classes
        _, _, self.vocab = get_word_emb_arr(VOCAB_DIR)
        if not input_dim:
            input_dim = len(self.vocab)
        self.input_dim = input_dim
        super(ClientModel, self).__init__(lr)

    def create_model(self):
        features = tf.placeholder(tf.float32, [None, self.input_dim])
        labels = tf.placeholder(tf.float32, [None, self.num_classes])
        
        W = tf.Variable(tf.random_normal(shape=[self.input_dim, self.num_classes]))
        b = tf.Variable(tf.random_normal(shape=[self.num_classes]))

        pred = tf.nn.softmax(tf.matmul(features, W) + b)

        # # Minimize error using cross entropy
        loss = tf.reduce_mean(-tf.reduce_sum(labels*tf.log(pred), reduction_indices=1))
        train_op = self.optimizer.minimize(
            loss=loss,
            global_step=tf.train.get_global_step())
        
        correct_pred = tf.equal(tf.argmax(pred,1), tf.argmax(labels,1))
        eval_metric_ops = tf.count_nonzero(correct_pred)
        
        return features, labels, train_op, eval_metric_ops

    def process_x(self, raw_x_batch):
        """
        Return:
            len(vocab) by len(raw_x_batch) np array
        """
        x_batch = [e[4] for e in raw_x_batch] # list of lines/phrases
        bags = [bag_of_words(line, self.vocab) for line in x_batch]
        bags = np.array(bags)
        return bags

    def process_y(self, raw_y_batch):
        y_batch = [int(e) for e in raw_y_batch]
        y_batch = [val_to_vec(self.num_classes, e) for e in y_batch]
        y_batch = np.array(y_batch)
        return y_batch





import argparse
import json

parser = argparse.ArgumentParser()

parser.add_argument('-f',
                help='path to .txt file containing word embedding information;',
                type=str,
                default='glove.6B.300d.txt')

args = parser.parse_args()

lines = []
with open(args.f, 'r') as inf:
    lines = inf.readlines()
lines = [l.split() for l in lines]
vocab = [l[0] for l in lines]
emb_floats = [[float(n) for n in l[1:]] for l in lines]
emb_floats.append([0.0 for _ in range(300)]) # for unknown word
js = {'vocab': vocab, 'emba': emb_floats}
with open('embs.json', 'w') as ouf:
    json.dump(js, ouf)



from pathlib import Path


from torch import nn, Tensor
from torch.utils.data import TensorDataset


from rayleaf.models.model import Model


IMAGE_SIZE = 28


class ClientModel(Model):
    def __init__(self, seed, lr, num_classes):
        super(ClientModel, self).__init__(seed, lr)

        self.num_classes = num_classes
        
        conv1 = nn.Conv2d(in_channels=1, out_channels=32, kernel_size=(5, 5), padding="same")
        pool1 = nn.MaxPool2d(kernel_size=(2,2), stride=2)
        conv2 = nn.Conv2d(in_channels=32, out_channels=64, kernel_size=(5, 5), padding="same")
        pool2 = nn.MaxPool2d(kernel_size=(2,2), stride=2)
        fc1 = nn.Linear(7 * 7 * 64, 2048)
        fc2 = nn.Linear(2048, self.num_classes)

        self.cnn = nn.Sequential(
            conv1,
            nn.ReLU(),
            pool1,
            conv2,
            nn.ReLU(),
            pool2,
            nn.Flatten(),
            fc1,
            nn.ReLU(),
            fc2
        )

        self.loss_fn = nn.CrossEntropyLoss()
        self.optimizer = self.optimizer(self.parameters(), lr=self.lr)
    
    
    def generate_dataset(self, data: dict, dataset_dir: Path) -> TensorDataset:
        return TensorDataset(
            Tensor(data["x"]).reshape(-1, 1, IMAGE_SIZE, IMAGE_SIZE),
            Tensor(data["y"]).long()
        )


    def forward(self, x):
        logits = self.cnn(x)

        return logits



from datetime import datetime


import rayleaf

from rayleaf.entities import Server, Client
from rayleaf.utils.logging_utils import log


from fedavg import fedavg


NUM_ROUNDS = 100
EVAL_EVERY = 1
NUM_CLIENTS = 10
CLIENTS_PER_ROUND = 5
BATCH_SIZE = 64
SEED = 0
NUM_EPOCHS = 25
GPUS_PER_CLIENT_CLUSTER = 1
NUM_CLIENT_CLUSTERS = 1
SAVE_MODEL = False
USE_GRADS = False
CPD_RANK = 10
RSVD_RANK = 16
ARCH_AWARE = True

OUTPUT = f"output/"
LR = 0.05

fedavg(
    dataset="femnist",
    output_dir= OUTPUT,
    num_rounds = NUM_ROUNDS,
    eval_every = EVAL_EVERY,
    num_clients = NUM_CLIENTS,
    clients_per_round = CLIENTS_PER_ROUND,
    client_lr = LR,
    batch_size = BATCH_SIZE,
    seed = SEED,
    num_epochs = NUM_EPOCHS,
    gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
    num_client_clusters = NUM_CLIENT_CLUSTERS,
    save_model = SAVE_MODEL
)

# for mu in [0.001, 0.01, 0.1, 1, 10]:
#     fedprox(
#         dataset="femnist",
#         output_dir= PAPER_OUTPUT,
#         mu=mu,
#         num_rounds = NUM_ROUNDS,
#         eval_every = EVAL_EVERY,
#         num_clients = NUM_CLIENTS,
#         clients_per_round = CLIENTS_PER_ROUND,
#         client_lr = FEMNIST_LR,
#         batch_size = BATCH_SIZE,
#         seed = SEED,
#         num_epochs = NUM_EPOCHS,
#         gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#         num_client_clusters = NUM_CLIENT_CLUSTERS,
#         save_model = SAVE_MODEL
#     )

# comp_bz2(
#     dataset="femnist",
#     output_dir= PAPER_OUTPUT,
#     num_rounds = NUM_ROUNDS,
#     eval_every = EVAL_EVERY,
#     num_clients = NUM_CLIENTS,
#     clients_per_round = CLIENTS_PER_ROUND,
#     client_lr = FEMNIST_LR,
#     batch_size = BATCH_SIZE,
#     seed = SEED,
#     num_epochs = NUM_EPOCHS,
#     gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#     num_client_clusters = NUM_CLIENT_CLUSTERS,
#     save_model = SAVE_MODEL
# )


# for stdev in [10 ** n for n in range(-4, 2)]:
#     for C in [10 ** n for n in range(-2, 4)]:
#         dpsgd_cnn(
#             stdev = stdev,
#             C = C,
#             num_rounds = NUM_ROUNDS,
#             eval_every = EVAL_EVERY,
#             num_clients = NUM_CLIENTS,
#             clients_per_round = CLIENTS_PER_ROUND,
#             client_lr = CLIENT_LR,
#             batch_size = BATCH_SIZE,
#             seed = SEED,
#             num_epochs = NUM_EPOCHS,
#             gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#             num_client_clusters = NUM_CLIENT_CLUSTERS,
#             save_model = SAVE_MODEL,
#             notes = f"stdev = {stdev}, C = {C}"
#         )

# sign_femnist(
#     num_rounds = NUM_ROUNDS,
#     eval_every = EVAL_EVERY,
#     num_clients = NUM_CLIENTS,
#     clients_per_round = CLIENTS_PER_ROUND,
#     client_lr = CLIENT_LR,
#     batch_size = BATCH_SIZE,
#     seed = SEED,
#     num_epochs = NUM_EPOCHS,
#     gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#     num_client_clusters = NUM_CLIENT_CLUSTERS,
#     save_model = SAVE_MODEL
# )

# for comp in ["rsvd", "svd", "qrcp"]:
#     for rank in [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]:
#         if comp == "rsvd" or (comp == "svd" and rank in (1, 2)):
#             continue
#         else:
#             nla_cnn(
#                 compression=comp,
#                 rank=rank,
#                 num_rounds = NUM_ROUNDS,
#                 eval_every = EVAL_EVERY,
#                 num_clients = NUM_CLIENTS,
#                 clients_per_round = CLIENTS_PER_ROUND,
#                 client_lr = CLIENT_LR,
#                 batch_size = BATCH_SIZE,
#                 seed = SEED,
#                 num_epochs = NUM_EPOCHS,
#                 gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#                 num_client_clusters = NUM_CLIENT_CLUSTERS,
#                 save_model = SAVE_MODEL
#             )

# for rank in [2 ** n for n in range(0, 11)]:
# nla_cnn(
#     output_dir=f"output/cpd/rsvd",
#     rank=RSVD_RANK,
#     arch_aware=ARCH_AWARE,
#     num_rounds = NUM_ROUNDS,
#     eval_every = EVAL_EVERY,
#     num_clients = NUM_CLIENTS,
#     clients_per_round = CLIENTS_PER_ROUND,
#     client_lr = CLIENT_LR,
#     batch_size = BATCH_SIZE,
#     seed = SEED,
#     num_epochs = NUM_EPOCHS,
#     gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#     num_client_clusters = NUM_CLIENT_CLUSTERS,
#     save_model = SAVE_MODEL
# )

# scaffold(
#     dataset="femnist",
#     global_lr="0.005",
#     num_rounds = NUM_ROUNDS,
#     eval_every = EVAL_EVERY,
#     num_clients = NUM_CLIENTS,
#     clients_per_round = CLIENTS_PER_ROUND,
#     client_lr = CLIENT_LR,
#     batch_size = BATCH_SIZE,
#     seed = SEED,
#     num_epochs = NUM_EPOCHS,
#     gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#     num_client_clusters = NUM_CLIENT_CLUSTERS,
#     save_model = SAVE_MODEL,
# )

# for cpd_rank in [1, 2, 4, 8, 16]:
#     cpd(
#         output_dir=f"output/cpd/cpd-layer0-r{cpd_rank}",
#         cpd_rank=cpd_rank,
#         rsvd_rank=RSVD_RANK,
#         num_rounds = NUM_ROUNDS,
#         eval_every = EVAL_EVERY,
#         num_clients = NUM_CLIENTS,
#         clients_per_round = CLIENTS_PER_ROUND,
#         client_lr = CLIENT_LR,
#         batch_size = BATCH_SIZE,
#         seed = SEED,
#         num_epochs = NUM_EPOCHS,
#         gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#         num_client_clusters = NUM_CLIENT_CLUSTERS,
#         save_model = SAVE_MODEL
#     )

# for rank in [32]:
#     try:
#         cpd(
#             output_dir=f"output/compression/conv1-r{rank}",
#             conv1_rank=rank,
#             num_rounds = NUM_ROUNDS,
#             eval_every = EVAL_EVERY,
#             num_clients = NUM_CLIENTS,
#             clients_per_round = CLIENTS_PER_ROUND,
#             client_lr = CLIENT_LR,
#             batch_size = BATCH_SIZE,
#             seed = SEED,
#             num_epochs = NUM_EPOCHS,
#             gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#             num_client_clusters = NUM_CLIENT_CLUSTERS,
#             save_model = SAVE_MODEL
#         )
#     except:
#         continue
# for rank in [1, 2, 4, 8, 16, 32, 64, 128]:
#     try:
#         cpd(
#             output_dir=f"output/compression/conv2-r{rank}",
#             conv2_rank=rank,
#             num_rounds = NUM_ROUNDS,
#             eval_every = EVAL_EVERY,
#             num_clients = NUM_CLIENTS,
#             clients_per_round = CLIENTS_PER_ROUND,
#             client_lr = CLIENT_LR,
#             batch_size = BATCH_SIZE,
#             seed = SEED,
#             num_epochs = NUM_EPOCHS,
#             gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#             num_client_clusters = NUM_CLIENT_CLUSTERS,
#             save_model = SAVE_MODEL
#         )
#     except:
#         continue
# for rank in [4]:
#     try:
#         cpd(
#             output_dir=f"output/compression/fc2-r{rank}",
#             fc2_rank=rank,
#             num_rounds = NUM_ROUNDS,
#             eval_every = EVAL_EVERY,
#             num_clients = NUM_CLIENTS,
#             clients_per_round = CLIENTS_PER_ROUND,
#             client_lr = CLIENT_LR,
#             batch_size = BATCH_SIZE,
#             seed = SEED,
#             num_epochs = NUM_EPOCHS,
#             gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#             num_client_clusters = NUM_CLIENT_CLUSTERS,
#             save_model = SAVE_MODEL
#         )
#     except:
#         continue

# cpd_femnist(
#     output_dir=f"output/compression/all_layers-r4-r16-r16-r8",
#     conv1_rank=4,
#     conv2_rank=16,
#     fc1_rank=16,
#     fc2_rank=8,
#     num_rounds = NUM_ROUNDS,
#     eval_every = EVAL_EVERY,
#     num_clients = NUM_CLIENTS,
#     clients_per_round = CLIENTS_PER_ROUND,
#     client_lr = CLIENT_LR,
#     batch_size = BATCH_SIZE,
#     seed = SEED,
#     num_epochs = NUM_EPOCHS,
#     gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#     num_client_clusters = NUM_CLIENT_CLUSTERS,
#     save_model = SAVE_MODEL
# )

# cpd_sc(
#     output_dir=f"output/cpd/sc/no_comp",
#     num_rounds = NUM_ROUNDS,
#     eval_every = EVAL_EVERY,
#     num_clients = NUM_CLIENTS,
#     clients_per_round = CLIENTS_PER_ROUND,
#     client_lr = CLIENT_LR,
#     batch_size = BATCH_SIZE,
#     seed = SEED,
#     num_epochs = NUM_EPOCHS,
#     gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#     num_client_clusters = NUM_CLIENT_CLUSTERS,
#     save_model = SAVE_MODEL
# )

# class PrintServer(Server):
#     def init(self):
#         log(f"Model layer shapes: {self.model_params.shapes}")


# rayleaf.run_experiment(
#     dataset = "shakespeare",
#     dataset_dir = "data/shakespeare/",
#     output_dir= f"output/shakespeare/{datetime.now()}/",
#     model = "stacked_lstm",
#     num_rounds = NUM_ROUNDS,
#     eval_every = EVAL_EVERY,
#     ServerType=PrintServer,
#     client_types=[(Client, NUM_CLIENTS)],
#     clients_per_round = CLIENTS_PER_ROUND,
#     client_lr = CLIENT_LR,
#     batch_size = BATCH_SIZE,
#     seed = SEED,
#     use_val_set = False,
#     num_epochs = NUM_EPOCHS,
#     gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#     num_client_clusters = NUM_CLIENT_CLUSTERS,
#     save_model = SAVE_MODEL
# )

# for r in [8, 16]:
# rsvd_shakespeare(
#     output_dir=f"output/shakespeare/compression/rsvd_r{8}_{datetime.now()}/",
#     # rank=r,
#     num_rounds = NUM_ROUNDS,
#     eval_every = EVAL_EVERY,
#     num_clients = NUM_CLIENTS,
#     clients_per_round = CLIENTS_PER_ROUND,
#     client_lr = CLIENT_LR,
#     batch_size = BATCH_SIZE,
#     seed = SEED,
#     num_epochs = NUM_EPOCHS,
#     gpus_per_client_cluster = GPUS_PER_CLIENT_CLUSTER,
#     num_client_clusters = NUM_CLIENT_CLUSTERS,
#     save_model = SAVE_MODEL
# )


import numpy as np
import os
import sys
import tensorflow as tf

from tensorflow.contrib import rnn

from rayleaf.models.model import Model
from utils.language_utils import line_to_indices, get_word_emb_arr, val_to_vec


VOCAB_DIR = 'sent140/embs.json'


class ClientModel(Model):

    def __init__(self, seed, lr, seq_len, num_classes, n_hidden, emb_arr=None):
        self.seq_len = seq_len
        self.num_classes = num_classes
        self.n_hidden = n_hidden
        _, self.indd, vocab = get_word_emb_arr(VOCAB_DIR)
        self.vocab_size = len(vocab)
        if emb_arr:
            self.emb_arr = emb_arr
        super(ClientModel, self).__init__(seed, lr)

    def create_model(self):
        features = tf.placeholder(tf.int32, [None, self.seq_len])
        embedding = tf.get_variable(
            'embedding', [self.vocab_size + 1, self.n_hidden], dtype=tf.float32)
        x = tf.cast(tf.nn.embedding_lookup(embedding, features), tf.float32)
        labels = tf.placeholder(tf.float32, [None, self.num_classes])
        
        stacked_lstm = rnn.MultiRNNCell(
            [rnn.BasicLSTMCell(self.n_hidden) for _ in range(2)])
        outputs, _ = tf.nn.dynamic_rnn(stacked_lstm, x, dtype=tf.float32)
        fc1 = tf.layers.dense(inputs=outputs[:, -1, :], units=128)
        pred = tf.layers.dense(inputs=fc1, units=self.num_classes)
        
        loss = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits_v2(logits=pred, labels=labels))
        train_op = self.optimizer.minimize(
            loss=loss,
            global_step=tf.train.get_global_step())
        
        correct_pred = tf.equal(tf.argmax(pred, 1), tf.argmax(labels, 1))
        eval_metric_ops = tf.count_nonzero(correct_pred)
        
        return features, labels, train_op, eval_metric_ops, loss

    def process_x(self, raw_x_batch, max_words=25):
        x_batch = [e[4] for e in raw_x_batch]
        x_batch = [line_to_indices(e, self.indd, max_words) for e in x_batch]
        x_batch = np.array(x_batch)
        return x_batch

    def process_y(self, raw_y_batch):
        y_batch = [int(e) for e in raw_y_batch]
        y_batch = [val_to_vec(self.num_classes, e) for e in y_batch]
        y_batch = np.array(y_batch)
        return y_batch



from pathlib import Path


import torchaudio
import torch.nn as nn
import torch.nn.functional as F

from torch.utils.data import Dataset


from rayleaf.models.model import Model
from rayleaf.models.speech_commands import utils as sc_utils


ORIGINAL_FREQ = 16000
RESAMPLE_FREQ = 8000


class ClientModel(Model):
    def __init__(self, seed, lr, num_classes):
        super(ClientModel, self).__init__(seed, lr)

        self.num_classes = num_classes

        self.conv1 = nn.Conv1d(in_channels=1, out_channels=32, kernel_size=80, stride=16)
        self.bn1 = nn.BatchNorm1d(num_features=32)
        self.pool1 = nn.MaxPool1d(kernel_size=4)
        self.conv2 = nn.Conv1d(in_channels=32, out_channels=32, kernel_size=3)
        self.bn2 = nn.BatchNorm1d(num_features=32)
        self.pool2 = nn.MaxPool1d(kernel_size=4)
        self.conv3 = nn.Conv1d(in_channels=32, out_channels=2 * 32, kernel_size=3)
        self.bn3 = nn.BatchNorm1d(num_features=2 * 32)
        self.pool3 = nn.MaxPool1d(kernel_size=4)
        self.conv4 = nn.Conv1d(in_channels=2 * 32, out_channels=2 * 32, kernel_size=3)
        self.bn4 = nn.BatchNorm1d(num_features=2 * 32)
        self.pool4 = nn.MaxPool1d(kernel_size=4)
        self.fc1 = nn.Linear(2 * 32, self.num_classes)

        self.loss_fn = nn.CrossEntropyLoss()
        self.optimizer = self.optimizer(self.parameters(), lr=self.lr)

        self.collate_fn = sc_utils.make_collate_fn(frequency=ORIGINAL_FREQ)
        self.resample = torchaudio.transforms.Resample(orig_freq=ORIGINAL_FREQ, new_freq=RESAMPLE_FREQ)


    def forward(self, x):
        x = self.resample.to(x.device)(x)

        x = self.conv1(x)
        x = self.bn1(x)
        x = F.relu(x)
        x = self.pool1(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = F.relu(x)
        x = self.pool2(x)
        x = self.conv3(x)
        x = self.bn3(x)
        x = F.relu(x)
        x = self.pool3(x)
        x = self.conv4(x)
        x = self.bn4(x)
        x = F.relu(x)
        x = self.pool4(x)
        x = F.avg_pool1d(x, x.shape[-1])
        x = x.permute(0, 2, 1)
        x = self.fc1(x)

        return x


    def generate_dataset(self, data: Dataset, dataset_dir: Path) -> Dataset:
        return data



from pathlib import Path
from typing import Union


import pandas as pd


import rayleaf.stats as stats


def graph(output_dir: Union[Path, str], eval_set: str):
    stats_dir = stats.STATS_DIR(output_dir)

    agg_train_stats = pd.read_csv(stats.STATS_CSV(stats_dir, "train", aggregate=True))
    agg_eval_stats = pd.read_csv(stats.STATS_CSV(stats_dir, eval_set, aggregate=True))

    rounds = agg_train_stats[stats.ROUND_NUMBER_KEY].values
    train_percentiles = _collect_percentile_cols(agg_train_stats, dataset="train")
    eval_percentiles = _collect_percentile_cols(agg_eval_stats, dataset=eval_set)

    graphs_dir = stats.GRAPHS_DIR(output_dir)
    stats.line_plot(
        output_dir=graphs_dir,
        title="Client Train Accuracies",
        x=rounds,
        xlabel="Round number",
        ylabel="Accuracy",
        marker=".",
        **train_percentiles
    )

    stats.line_plot(
        output_dir=graphs_dir,
        title=f"Client {eval_set.title()} Accuracies",
        x=rounds,
        xlabel="Round number",
        ylabel="Accuracy",
        marker=".",
        **eval_percentiles
    )

    averages = {
        "train": agg_train_stats[stats.AVERAGE_KEY],
        eval_set: agg_eval_stats[stats.AVERAGE_KEY]
    }
    stats.line_plot(
        output_dir=graphs_dir,
        title=f"Average Train vs {eval_set.title()} Accuracies",
        x=rounds,
        xlabel="Round number",
        ylabel="Accuracy",
        marker=".",
        **averages
    )


def _collect_percentile_cols(stats_df: pd.DataFrame, dataset: str):
    cols = {}
    for col in stats_df.columns:
        if col != stats.ROUND_NUMBER_KEY and col != stats.AVERAGE_KEY:
            cols[col] = stats_df[col].values
    
    return cols



import importlib
import os

from pathlib import Path


import ray

from tqdm import tqdm


import rayleaf.utils as utils
import rayleaf.utils.logging_utils as logging_utils
import rayleaf.models.speech_commands.initialize_data as sc_initialize_data

from rayleaf.core.client_cluster import make_client_cluster
from rayleaf.utils.data_utils import read_data
from rayleaf.entities.client import Client
from rayleaf.entities.server import Server
from rayleaf.models.model_constants import MODEL_SETTINGS


def log_experiment_details(**kwargs):
    logging_utils.log(utils.SECTION_STR.format("Experiment details"))
    for k, v in kwargs.items():
        if v is not None:
            logging_utils.log(f"{str(k)}: {str(v)}")


def initialize_resources(
    dataset_input: str,
    output_dir: str,
    model: str,
    client_lr: float,
    seed: int = 0,
):
    output_dir = Path(output_dir)
    if not output_dir.is_dir():
        os.makedirs(output_dir, exist_ok=True)
        
    experiment_log = open(Path(output_dir, "log.txt"), mode="w+")
    logging_utils.logging_file = experiment_log

    dataset = verify_dataset(dataset_input)
    dataset_input = dataset_input.lower()

    model_path = f"{dataset}.{model}"
    try:
        mod = importlib.import_module(f"rayleaf.models.{model_path}")
        ClientModel = getattr(mod, "ClientModel")
        model_settings = dict(MODEL_SETTINGS[model_path])
        model_settings["seed"] = seed
        model_settings["lr"] = client_lr
    except ModuleNotFoundError as e:
        logging_utils.log(f"Dataset {dataset} and model {model} is not a valid dataset-model pair.")
        logging_utils.shutdown_log()
        raise e

    return ClientModel, model_settings


def create_entities(
    device: str,
    num_client_clusters: int,
    gpus_per_client_cluster: float,
    seed: float,
    model_settings: dict,
    ClientModel: type,
    ServerType: type,
    client_types: list,
    dataset: str,
    dataset_dir: str,
    use_val_set: bool
):
    logging_utils.log(f"Spawning {num_client_clusters} ClientClusters using {device} device (this may take a while)")
    client_clusters = setup_client_clusters(
        num_client_clusters=num_client_clusters,
        gpus_per_client_cluster=gpus_per_client_cluster,
        seed=seed,
        device=device
    )

    # Create client model, and share params with server model
    client_model = ClientModel(**model_settings)

    # Create server
    verify_server_input(ServerType=ServerType)
    server = ServerType(model_params=list(client_model.parameters()), client_clusters=client_clusters)

    # Create clients
    clients = setup_clients(
        clients=client_types,
        client_clusters=client_clusters,
        dataset=dataset,
        dataset_dir=dataset_dir,
        model=ClientModel,
        model_settings=model_settings,
        seed=seed,
        use_val_set=use_val_set
    )

    return server, clients


def setup_client_clusters(num_client_clusters: int, gpus_per_client_cluster: int, seed: float, device: str = "cpu"):
    ClientCluster = make_client_cluster(num_gpus=gpus_per_client_cluster)

    if num_client_clusters < 1:
        return [ClientCluster.remote(id=0)]

    client_clusters = []
    
    for id in range(num_client_clusters):
        client_clusters.append(ClientCluster.remote(id=id, seed=seed, device=device))
    
    return client_clusters


def setup_clients(
    clients: list,
    client_clusters: list,
    dataset: str,
    dataset_dir: str,
    model: type,
    model_settings: tuple,
    seed: float,
    use_val_set: bool = False,
) -> list:
    """Instantiates clients based on given train and test data directories.

    Return:
        all_clients: list of Client objects.
    """
    dataset_dir = Path(dataset_dir)

    eval_set = "test" if not use_val_set else "val"
    train_data_dir = Path(dataset_dir, "data", "train")
    test_data_dir = Path(dataset_dir, "data", eval_set)

    if "speech_commands" in dataset:
        num_clients = verify_clients_input(clients=clients, max_num_clients=sc_initialize_data.MAX_NUMBER_OF_CLIENTS, dataset="speech_commands")
        users, groups, train_data, test_data = sc_initialize_data.federate_dataset(root=dataset_dir, num_clients=num_clients, seed=seed, iid="iid" in dataset)
    else:
        users, groups, train_data, test_data = read_data(train_data_dir, test_data_dir)
        verify_clients_input(clients=clients, max_num_clients=len(users), dataset=dataset)

    clients = create_clients(
        clients=clients,
        client_clusters=client_clusters,
        users=users,
        groups=groups,
        dataset_dir=dataset_dir,
        train_data=train_data,
        eval_data=test_data,
        model=model,
        model_settings=model_settings
    )

    return clients


def create_clients(
    clients: list,
    client_clusters: list,
    users: list,
    groups: list,
    dataset_dir: Path,
    train_data: dict,
    eval_data: dict, 
    model: type,
    model_settings: tuple
) -> list:
    if not groups or len(groups) == 0:
        groups = [[] for _ in users]

    num_client_clusters = len(client_clusters)

    client_creation_futures = []

    user_group_pairs = zip(users, groups)
    client_num = 0

    for ClientType, count in clients:
        for _ in range(count):
            u, g = next(user_group_pairs)
            future = client_clusters[client_num % num_client_clusters].add_client.remote(
                ClientType=ClientType,
                client_num=client_num,
                client_id=u,
                dataset_dir=dataset_dir,
                train_data=train_data[u],
                eval_data=eval_data[u],
                model=model,
                model_settings=model_settings,
                group=g
            )
            client_creation_futures.append(future)

            client_num += 1
    
    num_futures = len(client_creation_futures)
    clients = {}
    with tqdm(total=num_futures, leave=False, desc="Creating clients") as pbar:
        while len(client_creation_futures) > 0:
            complete, incomplete = ray.wait(client_creation_futures)

            for client_num, ClientType in ray.get(complete):
                clients[client_num] = ClientType
                pbar.update(1)

            client_creation_futures = incomplete

    return clients


def verify_dataset(dataset: str):
    for valid_dataset in utils.DATASETS:
        if valid_dataset in dataset:
            return valid_dataset

    raise ValueError(f"Dataset {dataset} is not a valid dataset.")


def verify_server_input(ServerType: type):
    assert isinstance(ServerType, type), f"ServerType {ServerType} is not a class."
    assert issubclass(ServerType, Server), f"ServerType {ServerType} is not a subclass of Server."


def verify_clients_input(clients: list, max_num_clients: int, dataset: str):
    prompt = "Each element in clients list must be a length-2 tuple of the form (Client subclass: type, count: int)."
    total = 0

    num_client_types = len(clients)
    for idx, pair in enumerate(clients):

        assert len(pair) == 2, f"{prompt} Got {pair} at index {idx} of clients list."

        assert isinstance(pair[0], type), f"{prompt} First element at index {idx} of clients list has type {type(pair[0])}."
        assert issubclass(pair[0], Client), f"{prompt} First element at index {idx} of clients list is not a subclass of Client."

        assert isinstance(pair[1], int), f"{prompt} Second element at index {idx} or clients list has type {type(pair[1])}."

        if idx < num_client_types - 1:
            assert pair[1] > 0, f"Each Client count except for last one must be greater than 0. Got count of {pair[1]} at index {idx} of clients list."
        else:
            assert pair[1] == -1 or pair[1] > 0, f"Last Client count must be greater than 0 or be -1. Got count of {pair[1]} at index {idx} of clients list."
            if pair[1] == -1:
                new_pair = (pair[0], max_num_clients - total)
                clients[idx] = new_pair
        
        total += pair[1]
    
    assert total <= max_num_clients, f"Max number of clients for {dataset} dataset is {max_num_clients} but clients list defines {total} clients."

    return total



import collections
import numpy as np
import os
import pickle
import sys
import tensorflow as tf

from tensorflow.contrib import rnn

from rayleaf.models.model import Model

VOCABULARY_PATH = '../data/reddit/vocab/reddit_vocab.pck'


# Code adapted from https://github.com/tensorflow/models/blob/master/tutorials/rnn/ptb/ptb_word_lm.py
# and https://r2rt.com/recurrent-neural-networks-in-tensorflow-iii-variable-length-sequences.html
class ClientModel(Model):
    def __init__(self, seed, lr, seq_len, n_hidden, num_layers,
        keep_prob=1.0, max_grad_norm=5, init_scale=0.1):

        self.seq_len = seq_len
        self.n_hidden = n_hidden
        self.num_layers = num_layers
        self.keep_prob = keep_prob
        self.max_grad_norm = max_grad_norm

        # initialize vocabulary
        self.vocab, self.vocab_size, self.unk_symbol, self.pad_symbol = self.load_vocab()

        self.initializer = tf.random_uniform_initializer(-init_scale, init_scale)

        super(ClientModel, self).__init__(seed, lr)

    def create_model(self):

        with tf.variable_scope('language_model', reuse=None, initializer=self.initializer):
            features = tf.placeholder(tf.int32, [None, self.seq_len], name='features')
            labels = tf.placeholder(tf.int32, [None, self.seq_len], name='labels')
            self.sequence_length_ph = tf.placeholder(tf.int32, [None], name='seq_len_ph')
            self.sequence_mask_ph = tf.placeholder(tf.float32, [None, self.seq_len], name='seq_mask_ph')

            self.batch_size = tf.shape(features)[0]

            # word embedding
            embedding = tf.get_variable(
                'embedding', [self.vocab_size, self.n_hidden], dtype=tf.float32)
            inputs = tf.nn.embedding_lookup(embedding, features)

            # LSTM
            output, state = self._build_rnn_graph(inputs) # TODO: check!

            # softmax
            with tf.variable_scope('softmax'):
                softmax_w = tf.get_variable(
                    'softmax_w', [self.n_hidden, self.vocab_size], dtype=tf.float32)
                softmax_b = tf.get_variable('softmax_b', [self.vocab_size], dtype=tf.float32)
            
            logits = tf.nn.xw_plus_b(output, softmax_w, softmax_b)

            # correct predictions
            labels_reshaped = tf.reshape(labels, [-1])
            pred = tf.cast(tf.argmax(logits, 1), tf.int32)
            correct_pred = tf.cast(tf.equal(pred, labels_reshaped), tf.int32)
            
            # predicting unknown is always considered wrong
            unk_tensor = tf.fill(tf.shape(labels_reshaped), self.unk_symbol)
            pred_unk = tf.cast(tf.equal(pred, unk_tensor), tf.int32)
            correct_unk = tf.multiply(pred_unk, correct_pred)

            # predicting padding is always considered wrong
            pad_tensor = tf.fill(tf.shape(labels_reshaped), 0)
            pred_pad = tf.cast(tf.equal(pred, pad_tensor), tf.int32)
            correct_pad = tf.multiply(pred_pad, correct_pred)

            # Reshape logits to be a 3-D tensor for sequence loss
            logits = tf.reshape(logits, [-1, self.seq_len, self.vocab_size])

            # Use the contrib sequence loss and average over the batches
            loss = tf.contrib.seq2seq.sequence_loss(
                logits,
                labels,
                weights=self.sequence_mask_ph,
                average_across_timesteps=False,
                average_across_batch=True)

            # Update the cost
            #self.cost = tf.reduce_sum(loss)
            self.cost = tf.reduce_mean(loss)
            self.final_state = state

            tvars = tf.trainable_variables()
            grads, _ = tf.clip_by_global_norm(tf.gradients(self.cost, tvars), self.max_grad_norm)
            train_op = self.optimizer.apply_gradients(
                zip(grads, tvars),
                global_step=tf.train.get_or_create_global_step())

            eval_metric_ops = tf.count_nonzero(correct_pred) - tf.count_nonzero(correct_unk) - tf.count_nonzero(correct_pad)

        return features, labels, train_op, eval_metric_ops, self.cost

    def _build_rnn_graph(self, inputs):
        def make_cell():
            cell = tf.contrib.rnn.LSTMBlockCell(self.n_hidden, forget_bias=0.0)
            if self.keep_prob < 1:
                cell = tf.nn.rnn_cell.DropoutWrapper(cell, output_keep_prob=self.keep_prob)
            return cell

        cell = tf.nn.rnn_cell.MultiRNNCell(
            [make_cell() for _ in range(self.num_layers)], state_is_tuple=True)

        self.initial_state = cell.zero_state(self.batch_size, tf.float32)
        outputs, state = tf.nn.dynamic_rnn(
            cell, inputs, initial_state=self.initial_state, sequence_length=self.sequence_length_ph)
        output = tf.reshape(tf.concat(outputs, 1), [-1, self.n_hidden])
        return output, state

    def process_x(self, raw_x_batch):
        tokens = self._tokens_to_ids([s for s in raw_x_batch])
        lengths = np.sum(tokens != self.pad_symbol, axis=1)
        return tokens, lengths

    def process_y(self, raw_y_batch):
        tokens = self._tokens_to_ids([s for s in raw_y_batch])
        return tokens

    def _tokens_to_ids(self, raw_batch):
        def tokens_to_word_ids(tokens, vocab):
            return [vocab[word] for word in tokens]

        to_ret = [tokens_to_word_ids(seq, self.vocab) for seq in raw_batch]
        return np.array(to_ret)

    def batch_data(self, data, batch_size):
        data_x = data['x']
        data_y = data['y']

        perm = np.random.permutation(len(data['x']))
        data_x = [data_x[i] for i in perm]
        data_y = [data_y[i] for i in perm]

        # flatten lists
        def flatten_lists(data_x_by_comment, data_y_by_comment):
            data_x_by_seq, data_y_by_seq, mask_by_seq = [], [], []
            for c, l in zip(data_x_by_comment, data_y_by_comment):
                data_x_by_seq.extend(c)
                data_y_by_seq.extend(l['target_tokens'])
                mask_by_seq.extend(l['count_tokens'])

            if len(data_x_by_seq) % batch_size != 0:
                dummy_tokens = [self.pad_symbol for _ in range(self.seq_len)]
                dummy_mask = [0 for _ in range(self.seq_len)]
                num_dummy = batch_size - len(data_x_by_seq) % batch_size

                data_x_by_seq.extend([dummy_tokens for _ in range(num_dummy)])
                data_y_by_seq.extend([dummy_tokens for _ in range(num_dummy)])
                mask_by_seq.extend([dummy_mask for _ in range(num_dummy)])

            return data_x_by_seq, data_y_by_seq, mask_by_seq
        
        data_x, data_y, data_mask = flatten_lists(data_x, data_y)

        for i in range(0, len(data_x), batch_size):
            batched_x = data_x[i:i+batch_size]
            batched_y = data_y[i:i+batch_size]
            batched_mask = data_mask[i:i+batch_size]

            input_data, input_lengths = self.process_x(batched_x)
            target_data = self.process_y(batched_y)

            yield (input_data, target_data, input_lengths, batched_mask)

    def _run_epoch(self, data, batch_size=5):
        state = None

        fetches = {
            'cost': self.cost,
            'final_state': self.final_state,
        }

        for input_data, target_data, input_lengths, input_mask in self.batch_data(data, batch_size):

            feed_dict = {
                self.features: input_data,
                self.labels: target_data,
                self.sequence_length_ph: input_lengths,
                self.sequence_mask_ph: input_mask,
            }

            # We need to feed the input data so that the batch size can be inferred.
            if state is None:
                state = self.sess.run(self.initial_state, feed_dict=feed_dict)

            for i, (c, h) in enumerate(self.initial_state):
                feed_dict[c] = state[i].c
                feed_dict[h] = state[i].h

            with self.graph.as_default():
                _, vals = self.sess.run([self.train_op, fetches], feed_dict=feed_dict)
            
            state = vals['final_state']

    def test(self, data, batch_size=5):
        tot_acc, tot_samples = 0, 0
        tot_loss, tot_batches = 0, 0

        for input_data, target_data, input_lengths, input_mask in self.batch_data(data, batch_size):

            with self.graph.as_default():
                acc, loss = self.sess.run(
                    [self.eval_metric_ops, self.loss], 
                    feed_dict={
                        self.features: input_data,
                        self.labels: target_data,
                        self.sequence_length_ph: input_lengths, 
                        self.sequence_mask_ph: input_mask,
                    })
            
            tot_acc += acc
            tot_samples += np.sum(input_lengths)

            tot_loss += loss
            tot_batches += 1

        acc = float(tot_acc) / tot_samples # this top 1 accuracy considers every pred. of unknown and padding as wrong
        loss = tot_loss / tot_batches # the loss is already averaged over samples
        return {'accuracy': acc, 'loss': loss}

    def load_vocab(self):
        vocab_file = pickle.load(open(VOCABULARY_PATH, 'rb'))
        vocab = collections.defaultdict(lambda: vocab_file['unk_symbol'])
        vocab.update(vocab_file['vocab'])

        return vocab, vocab_file['size'], vocab_file['unk_symbol'], vocab_file['pad_symbol']




import random

from datetime import datetime


import numpy as np
import ray
import torch


import rayleaf.stage
import rayleaf.utils as utils
import rayleaf.utils.logging_utils as logging_utils

from rayleaf.utils.client_count_utils import client_counts_string, count_selected_clients, online


NUM_GPUS = torch.cuda.device_count()


def run_experiment(
    dataset: str,
    dataset_dir: str,
    output_dir: str,
    model: str,
    num_rounds: int,
    eval_every,
    ServerType: type,
    client_types: list,
    clients_per_round: int,
    client_lr: float,
    batch_size: int = 10,
    seed: int = 0,
    use_val_set: bool = False,
    num_epochs: int = 1,
    gpus_per_client_cluster: float = 1,
    num_client_clusters: int = NUM_GPUS,
    save_model: bool = False,
    notes: str = None
) -> None:
    start_time = datetime.now()
    
    ClientModel, model_settings = rayleaf.stage.initialize_resources(
        dataset_input=dataset,
        output_dir=output_dir,
        model=model,
        client_lr=client_lr,
        seed=seed,
    )

    rayleaf.stage.log_experiment_details(
        dataset=dataset,
        dataset_dir=dataset_dir,
        output_dir=output_dir,
        model=model,
        num_rounds=num_rounds,
        eval_every=eval_every,
        ServerType=ServerType,
        client_types=client_types,
        clients_per_round=clients_per_round,
        client_lr=client_lr,
        batch_size=batch_size,
        seed=seed,
        use_val_set=use_val_set,
        num_epochs=num_epochs,
        gpus_per_client_cluster=gpus_per_client_cluster,
        num_client_clusters=num_client_clusters,
        save_model=save_model,
        notes = notes
    )

    logging_utils.log(utils.SECTION_STR.format("Simulation"))
    ray.init()

    # Set the random seed if provided (affects client sampling, and batching)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)

    # Get cpu or gpu device for training.
    device = "cuda" if torch.cuda.is_available() else "cpu"

    server, clients = rayleaf.stage.create_entities(
        device=device,
        num_client_clusters=num_client_clusters,
        gpus_per_client_cluster=gpus_per_client_cluster,
        seed=seed,
        model_settings=model_settings,
        ClientModel=ClientModel,
        ServerType=ServerType,
        client_types=client_types,
        dataset=dataset,
        dataset_dir=dataset_dir,
        use_val_set=use_val_set
    )

    if clients_per_round == -1:
        clients_per_round = len(clients)

    client_counts = count_selected_clients(online(clients), clients)
    logging_utils.log(f"{len(clients)} total clients: {client_counts_string(client_counts)}")
    logging_utils.log()

    eval_set = "test" if not use_val_set else "val"
    rayleaf.stage.eval_server(
        server=server,
        num_round=0,
        eval_set=eval_set,
        output_dir=output_dir,
        batch_size=batch_size
    )

    if type(eval_every) == int:
        eval_every = set(range(0, num_rounds + 1, eval_every))
    eval_every = set(eval_every)

    # Simulate training
    for round_number in range(1, num_rounds + 1):
        rayleaf.stage.train(
            server=server,
            clients=clients,
            clients_per_round=clients_per_round,
            output_dir=output_dir,
            num_epochs=num_epochs,
            batch_size=batch_size,
            curr_round=round_number,
            num_rounds=num_rounds,
            log_progress=True
        )

        # Test model
        if round_number in eval_every:
            rayleaf.stage.eval_server(
                server=server,
                num_round=round_number,
                eval_set=eval_set,
                output_dir=output_dir,
                batch_size=batch_size
            )
    
    logging_utils.log(utils.SECTION_STR.format("Post-Simulation"))
        
    rayleaf.stage.graph(
        output_dir=output_dir,
        eval_set=eval_set
    )

    rayleaf.stage.teardown(
        save_model=save_model,
        output_dir=output_dir,
        dataset=dataset,
        model=model,
        server=server,
        start_time=start_time
    )

    ray.shutdown()



from pathlib import Path


from torch import nn, Tensor
from torch.utils.data import TensorDataset


from rayleaf.models.model import Model
from rayleaf.utils.language_utils import letter_to_index, word_to_indices


class ClientModel(Model):
    def __init__(self, seed, lr, seq_len, num_classes, num_hidden):
        super(ClientModel, self).__init__(seed, lr)

        self.seq_len = seq_len
        self.num_classes = num_classes
        self.num_hidden = num_hidden

        self.embedding = nn.Embedding(self.num_classes, 8)
        self.lstm = nn.LSTM(input_size=8, hidden_size=self.num_hidden, num_layers=2, batch_first=True)
        self.fc1 = nn.Linear(self.num_hidden * 2, self.num_classes)

        self.loss_fn = nn.CrossEntropyLoss()
        self.optimizer = self.optimizer(self.parameters(), lr=self.lr)


    def generate_dataset(self, data: dict, dataset_dir: Path) -> TensorDataset:
        return TensorDataset(
            Tensor([word_to_indices(x) for x in data["x"]]).long(),
            Tensor([letter_to_index(y) for y in data["y"]]).long()
        )


    def forward(self, x):
        embedding = self.embedding(x)
        _, (hn, _) = self.lstm(embedding)
        hn = hn.transpose(0, 1).reshape(-1, 2 * self.num_hidden)
        logits = self.fc1(hn)

        return logits



import numpy as np
import torch

from scipy.linalg import block_diag
from torch import nn, Tensor


from .utils import _tensor_to_block_diag, _block_diag_to_tensor


HANDLED_FUNCTIONS = {}

def implements(np_function):
   "Register an __array_function__ implementation for DiagonalArray objects."
   def decorator(func):
       HANDLED_FUNCTIONS[np_function] = func
       return func
   return decorator


class TensorArray(np.lib.mixins.NDArrayOperatorsMixin):


    def _find_shapes(arg) -> tuple:
        try:
            for iterable in arg:
                if not isinstance(iterable, torch.Size):
                    for element in iterable:
                        if not (isinstance(element, int) and element > 0):
                            raise ValueError(f"Shape iterable {arg} contains non-integer values or integers less than"
                                " or equal to 0")
            return arg
        except TypeError:
            try:
                return arg.shapes
            except AttributeError:
                raise TypeError(f"Input argument must be an iterable, Model, or TensorArray but found {type(arg)}")


    def _tensor_from_shapes(shapes, tensor_generator, device):
        shapes = TensorArray._find_shapes(shapes)
        tensors = []

        for shape in shapes:
            tensors.append(tensor_generator(shape).to(device))

        return TensorArray(tensors)


    def zeros(arg, device="cpu"):
        return TensorArray._tensor_from_shapes(shapes=arg, tensor_generator=torch.zeros, device=device)


    def ones(arg, device="cpu"):
        return TensorArray._tensor_from_shapes(shapes=arg, tensor_generator=torch.ones, device=device)


    def randn(arg, device="cpu"):
        return TensorArray._tensor_from_shapes(shapes=arg, tensor_generator=torch.randn, device=device)


    def __init__(self, tensors, device="cpu") -> None:
        if isinstance(tensors, nn.Module):
            tensors = tensors.parameters()

        tensors = list(tensors)

        if len(tensors) > 0:
            for t in tensors:
                if isinstance(t, Tensor):
                    device = t.device
                    break

        try:
            self._tensors = tuple(t.detach().clone().to(device) if isinstance(t, Tensor) \
                else Tensor(t).requires_grad_(False).to(device) for t in tensors)
        except TypeError:
            raise TypeError(f"Cannot create TensorArray from object that is not iterable and"
                " not an instance of torch.nn.Module. Found object of type {type(tensors)}")

        self._dtype = self.tensors[0].dtype if len(self.tensors) > 0 else None
        
        self._shapes = None
        self._size = None


    @property
    def tensors(self) -> tuple:
        return self._tensors


    @property
    def dtype(self):
        return self._dtype


    def _compute_shapes_and_size(self) -> None:
        self._shapes = []
        self._size = 0
        for param in self.tensors:
            self._shapes.append(param.shape)
            self._size += np.prod(param.shape)
        self._shapes = tuple(self._shapes)


    @property
    def shapes(self) -> tuple:
        if self._shapes is None:
            self._compute_shapes_and_size()
        
        return self._shapes


    @property
    def size(self) -> int:
        if self._size is None:
            self._compute_shapes_and_size()
        
        return self._size


    def __repr__(self):
        repr = f"{self.__class__.__name__}(params={self._tensors})"
        return repr


    def __str__(self):
        st = f"Params with shape {self.shapes}\n"
        for i, param in enumerate(self.tensors):
            st += f"Param {i}: {param}\n"
        return st


    def __len__(self):
        return len(self.shapes)


    def __iadd__(self, other):
        return self + other


    def __isub__(self, other):
        return self - other


    def __imul__(self, other):
        return self * other


    def __itruediv__(self, other):
        return self / other


    def __ifloordiv__(self, other):
        return self // other


    def __imod__(self, other):
        return self % other


    def __ipow__(self, other):
        return self ** other


    def _compare(self, other, operator, default):
        if type(self) != type(other):
            return default
        elif len(self) != len(other):
            return default
        elif self.shapes != other.shapes:
            return default
        
        res = []
        for i in range(len(self)):
            res.append(operator(self.tensors[i], other.tensors[i]))
        
        return self.__class__(res)


    def __eq__(self, other):
        return self._compare(other, lambda x, y: x == y, default=False)


    def __ne__(self, other):
        return self._compare(other, lambda x, y: x != y, default=True)

    
    def __array__(self, dtype=None) -> Tensor:
        flattened_layers = []
        for tensor in self._tensors:
            flattened_layers.append(tensor.flatten())
        flattened = torch.cat(flattened_layers)
        
        return flattened.cpu().to(dtype)


    def flat(self, dtype=None):
        return self.__array__(dtype=dtype)


    def unflatten(flat_arr, shapes):
        try:
            if (flat_arr.ndim > 1):
                raise TypeError(f"Input array should be 1-dimensional but has shape {flat_arr.shape}")
        except AttributeError:
            raise TypeError(f"Input array should be a numpy.ndarray or torch.Tensor but found {type(flat_arr)}")

        output_tensors = []
        start_index = 0

        for shape in shapes:
            try:
                count = np.prod(shape)
                output_tensors.append(Tensor(flat_arr[start_index:start_index + count]).reshape(shape).requires_grad_(False))
            except TypeError:
                raise TypeError(f"shapes parameter must be an iterable of iterables of ints")
            start_index += count

        if start_index != flat_arr.shape[0]:
            raise TypeError(f"Input array length and shapes mismatch. Input array has length {flat_arr.shape[0]}"
                " but shapes implies length of {start_index}")
        
        return TensorArray(output_tensors)


    def as_block_diag(self, dtype=None):
        mats = []
        for tensor in self.tensors:
            mats.append(_tensor_to_block_diag(tensor))
        
        blocked = block_diag(*mats)
        return Tensor(blocked).to(dtype).requires_grad_(False)


    def from_block_diag(mat, shapes):
        curr_inds = [0, 0]

        tensors = []
        for shape in shapes:
            tensors.append(_block_diag_to_tensor(mat, shape, start_inds=curr_inds))
        
        return TensorArray(tensors)


    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method == '__call__':
            first_shapes = None

            sanitized_inputs = []
            for input_ in inputs:
                if isinstance(input_, self.__class__):
                    if first_shapes is None:
                        first_shapes = input_.shapes
                    elif first_shapes != input_.shapes:
                        raise TypeError(f"Found mismatched dimensions: {first_shapes} and {input_.shapes}")

                    sanitized_inputs.append(input_.flat())
                else:
                    sanitized_inputs.append(input_)

            flat_output = ufunc(*sanitized_inputs, **kwargs)
                    
            return self.__class__.unflatten(flat_output, first_shapes)
        else:
            return NotImplemented

    def __array_function__(self, func, types, args, kwargs):
        if func not in HANDLED_FUNCTIONS:
            return NotImplemented
        # Note: this allows subclasses that don't override
        # __array_function__ to handle DiagonalArray objects.
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        return HANDLED_FUNCTIONS[func](*args, **kwargs)


    @implements(np.linalg.norm)
    def norm(x, ord=None, axis=None, keepdims=False):
        return np.linalg.norm(x.flat(), ord=ord, axis=axis, keepdims=keepdims)


    def to(self, arg):
        new_tensors = [t.to(arg) for t in self.tensors]
        return TensorArray(new_tensors)



"""
Based on https://pytorch.org/tutorials/intermediate/speech_command_classification_with_torchaudio_tutorial.html#formatting-the-data
"""


import torch
import torch.nn.functional as F


LABELS = [
    "backward",
    "bed",
    "bird",
    "cat",
    "dog",
    "down",
    "eight",
    "five",
    "follow",
    "forward",
    "four",
    "go",
    "happy",
    "house",
    "learn",
    "left",
    "marvin",
    "nine",
    "no",
    "off",
    "on",
    "one",
    "right",
    "seven",
    "sheila",
    "six",
    "stop",
    "three",
    "tree",
    "two",
    "up",
    "visual",
    "wow",
    "yes",
    "zero"
]


def label_to_index(word):
    return torch.tensor(LABELS.index(word))


def index_to_label(index):
    return LABELS[index]


def pad_waveform(waveform, frequency: int):
    return F.pad(waveform, (0, frequency - waveform.shape[-1]), mode="constant", value=0.0)


def make_collate_fn(frequency: int):
    def collate_fn(batch):
        tensors, targets = [], []

        for waveform, _, label, *_ in batch:
            tensors.append(pad_waveform(waveform, frequency))
            targets.append(label_to_index(label))

        tensors = torch.stack(tensors)
        targets = torch.stack(targets)

        return tensors, targets
    
    return collate_fn



import random

from pathlib import Path


import numpy as np
import ray
import torch


def make_client_cluster(num_gpus: float) -> type:

    @ray.remote(num_gpus=num_gpus)
    class ClientCluster:
        def __init__(self, id: int, seed: float, device: str = "cpu"):
            self.seed = seed
            self.set_seeds(seed)

            self.id = id
            self.device = device
            self.clients = {}


        def add_client(
            self,
            ClientType: type,
            client_num: int,
            client_id: str,
            dataset_dir: Path,
            train_data: dict,
            eval_data: dict,
            model: type,
            model_settings: tuple,
            group: list = None
        ) -> None:
            self.clients[client_num] = ClientType(
                client_num,
                client_id,
                dataset_dir,
                train_data,
                eval_data,
                model,
                model_settings,
                group,
                device=self.device
            )

            return client_num, ClientType

        
        def train_clients(
            self,
            server_update,
            selected_clients: list = None,
            num_epochs: int = 1,
            batch_size: int = 10
        ) -> list:

            clients_to_train = self.get_clients_from_client_nums(selected_clients)
            
            updates = []

            for client in clients_to_train:
                training_result = client._train(server_update, num_epochs, batch_size)

                updates.append(training_result)
            
            return updates


        def eval_model(
            self,
            model_params: list,
            selected_clients: list = None,
            set_to_use: str = "test",
            batch_size: int = 10
        ) -> dict:
            clients_to_eval = self.get_clients_from_client_nums(selected_clients)

            stats = []

            for client in clients_to_eval:
                client.model_params = model_params
                client_stats = client._eval(set_to_use, batch_size)
                stats.append(client_stats)
            
            return stats


        def get_clients_info(self, client_nums: list = None) -> tuple:
            clients = self.get_clients_from_client_nums(client_nums)

            ids = []
            groups = {}
            num_samples = {}
            for client in clients:
                ids.append(client.id)
                groups[client.id] = client.group
                num_samples[client.id] = client.num_samples

            return ids, groups, num_samples


        def get_clients_from_client_nums(self, client_nums: list) -> list:
            if client_nums is None:
                return list(self.clients.values())

            clients = []

            for client_num in client_nums:
                clients.append(self.clients[client_num])

            return clients


        def set_seeds(self, seed: float = 0):
            random.seed(seed)
            np.random.seed(seed)
            torch.manual_seed(seed)
            torch.cuda.manual_seed(seed)
    
    return ClientCluster
